from PIL import Image

from .helper import PillowTestCase, unittest

TEST_FILE = "Tests/images/fli_overflow.fli"


class TestFliOverflow(PillowTestCase):
    def test_fli_overflow(self):

        # this should not crash with a malloc error or access violation
        im = Image.open(TEST_FILE)
        im.load()


if __name__ == "__main__":
    unittest.main()



from PIL import GbrImagePlugin, Image

from .helper import PillowTestCase


class TestFileGbr(PillowTestCase):
    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, GbrImagePlugin.GbrImageFile, invalid_file)

    def test_gbr_file(self):
        im = Image.open("Tests/images/gbr.gbr")

        target = Image.open("Tests/images/gbr.png")

        self.assert_image_equal(target, im)



from PIL._util import py3

from .helper import PillowTestCase, hopper


class TestImageGetIm(PillowTestCase):
    def test_sanity(self):
        im = hopper()
        type_repr = repr(type(im.getim()))

        if py3:
            self.assertIn("PyCapsule", type_repr)

        self.assertIsInstance(im.im.id, int)



from PIL import ImageQt

from .helper import PillowTestCase, hopper
from .test_imageqt import PillowQPixmapTestCase


class TestFromQPixmap(PillowQPixmapTestCase, PillowTestCase):
    def roundtrip(self, expected):
        result = ImageQt.fromqpixmap(ImageQt.toqpixmap(expected))
        # Qt saves all pixmaps as rgb
        self.assert_image_equal(result, expected.convert("RGB"))

    def test_sanity(self):
        for mode in ("1", "RGB", "RGBA", "L", "P"):
            self.roundtrip(hopper(mode))



from __future__ import print_function

import glob
import os
import sys
import traceback

sys.path.insert(0, ".")

for file in glob.glob("src/PIL/*.py"):
    module = os.path.basename(file)[:-3]
    try:
        exec("from PIL import " + module)
    except (ImportError, SyntaxError):
        print("===", "failed to import", module)
        traceback.print_exc()



from PIL import ImagePalette

from .helper import PillowTestCase, hopper


class TestImagePutPalette(PillowTestCase):
    def test_putpalette(self):
        def palette(mode):
            im = hopper(mode).copy()
            im.putpalette(list(range(256)) * 3)
            p = im.getpalette()
            if p:
                return im.mode, p[:10]
            return im.mode

        self.assertRaises(ValueError, palette, "1")
        for mode in ["L", "LA", "P", "PA"]:
            self.assertEqual(
                palette(mode),
                ("PA" if "A" in mode else "P", [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
            )
        self.assertRaises(ValueError, palette, "I")
        self.assertRaises(ValueError, palette, "F")
        self.assertRaises(ValueError, palette, "RGB")
        self.assertRaises(ValueError, palette, "RGBA")
        self.assertRaises(ValueError, palette, "YCbCr")

    def test_imagepalette(self):
        im = hopper("P")
        im.putpalette(ImagePalette.negative())
        im.putpalette(ImagePalette.random())
        im.putpalette(ImagePalette.sepia())
        im.putpalette(ImagePalette.wedge())



from io import BytesIO

from PIL import EpsImagePlugin, Image, ImageFile

from .helper import PillowTestCase, fromstring, hopper, tostring, unittest

try:
    from PIL import _webp

    HAVE_WEBP = True
except ImportError:
    HAVE_WEBP = False


codecs = dir(Image.core)

# save original block sizes
MAXBLOCK = ImageFile.MAXBLOCK
SAFEBLOCK = ImageFile.SAFEBLOCK


class TestImageFile(PillowTestCase):
    def test_parser(self):
        def roundtrip(format):

            im = hopper("L").resize((1000, 1000))
            if format in ("MSP", "XBM"):
                im = im.convert("1")

            test_file = BytesIO()

            im.copy().save(test_file, format)

            data = test_file.getvalue()

            parser = ImageFile.Parser()
            parser.feed(data)
            imOut = parser.close()

            return im, imOut

        self.assert_image_equal(*roundtrip("BMP"))
        im1, im2 = roundtrip("GIF")
        self.assert_image_similar(im1.convert("P"), im2, 1)
        self.assert_image_equal(*roundtrip("IM"))
        self.assert_image_equal(*roundtrip("MSP"))
        if "zip_encoder" in codecs:
            try:
                # force multiple blocks in PNG driver
                ImageFile.MAXBLOCK = 8192
                self.assert_image_equal(*roundtrip("PNG"))
            finally:
                ImageFile.MAXBLOCK = MAXBLOCK
        self.assert_image_equal(*roundtrip("PPM"))
        self.assert_image_equal(*roundtrip("TIFF"))
        self.assert_image_equal(*roundtrip("XBM"))
        self.assert_image_equal(*roundtrip("TGA"))
        self.assert_image_equal(*roundtrip("PCX"))

        if EpsImagePlugin.has_ghostscript():
            im1, im2 = roundtrip("EPS")
            # This test fails on Ubuntu 12.04, PPC (Bigendian) It
            # appears to be a ghostscript 9.05 bug, since the
            # ghostscript rendering is wonky and the file is identical
            # to that written on ubuntu 12.04 x64
            # md5sum: ba974835ff2d6f3f2fd0053a23521d4a

            # EPS comes back in RGB:
            self.assert_image_similar(im1, im2.convert("L"), 20)

        if "jpeg_encoder" in codecs:
            im1, im2 = roundtrip("JPEG")  # lossy compression
            self.assert_image(im1, im2.mode, im2.size)

        self.assertRaises(IOError, roundtrip, "PDF")

    def test_ico(self):
        with open("Tests/images/python.ico", "rb") as f:
            data = f.read()
        with ImageFile.Parser() as p:
            p.feed(data)
            self.assertEqual((48, 48), p.image.size)

    def test_safeblock(self):
        if "zip_encoder" not in codecs:
            self.skipTest("PNG (zlib) encoder not available")

        im1 = hopper()

        try:
            ImageFile.SAFEBLOCK = 1
            im2 = fromstring(tostring(im1, "PNG"))
        finally:
            ImageFile.SAFEBLOCK = SAFEBLOCK

        self.assert_image_equal(im1, im2)

    def test_raise_ioerror(self):
        self.assertRaises(IOError, ImageFile.raise_ioerror, 1)

    def test_raise_typeerror(self):
        with self.assertRaises(TypeError):
            parser = ImageFile.Parser()
            parser.feed(1)

    def test_negative_stride(self):
        with open("Tests/images/raw_negative_stride.bin", "rb") as f:
            input = f.read()
        p = ImageFile.Parser()
        p.feed(input)
        with self.assertRaises(IOError):
            p.close()

    def test_truncated_with_errors(self):
        if "zip_encoder" not in codecs:
            self.skipTest("PNG (zlib) encoder not available")

        im = Image.open("Tests/images/truncated_image.png")
        with self.assertRaises(IOError):
            im.load()

        # Test that the error is raised if loaded a second time
        with self.assertRaises(IOError):
            im.load()

    def test_truncated_without_errors(self):
        if "zip_encoder" not in codecs:
            self.skipTest("PNG (zlib) encoder not available")

        im = Image.open("Tests/images/truncated_image.png")

        ImageFile.LOAD_TRUNCATED_IMAGES = True
        try:
            im.load()
        finally:
            ImageFile.LOAD_TRUNCATED_IMAGES = False

    def test_broken_datastream_with_errors(self):
        if "zip_encoder" not in codecs:
            self.skipTest("PNG (zlib) encoder not available")

        im = Image.open("Tests/images/broken_data_stream.png")
        with self.assertRaises(IOError):
            im.load()

    def test_broken_datastream_without_errors(self):
        if "zip_encoder" not in codecs:
            self.skipTest("PNG (zlib) encoder not available")

        im = Image.open("Tests/images/broken_data_stream.png")

        ImageFile.LOAD_TRUNCATED_IMAGES = True
        try:
            im.load()
        finally:
            ImageFile.LOAD_TRUNCATED_IMAGES = False


class MockPyDecoder(ImageFile.PyDecoder):
    def decode(self, buffer):
        # eof
        return -1, 0


xoff, yoff, xsize, ysize = 10, 20, 100, 100


class MockImageFile(ImageFile.ImageFile):
    def _open(self):
        self.rawmode = "RGBA"
        self.mode = "RGBA"
        self._size = (200, 200)
        self.tile = [("MOCK", (xoff, yoff, xoff + xsize, yoff + ysize), 32, None)]


class TestPyDecoder(PillowTestCase):
    def get_decoder(self):
        decoder = MockPyDecoder(None)

        def closure(mode, *args):
            decoder.__init__(mode, *args)
            return decoder

        Image.register_decoder("MOCK", closure)
        return decoder

    def test_setimage(self):
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)
        d = self.get_decoder()

        im.load()

        self.assertEqual(d.state.xoff, xoff)
        self.assertEqual(d.state.yoff, yoff)
        self.assertEqual(d.state.xsize, xsize)
        self.assertEqual(d.state.ysize, ysize)

        self.assertRaises(ValueError, d.set_as_raw, b"\x00")

    def test_extents_none(self):
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)
        im.tile = [("MOCK", None, 32, None)]
        d = self.get_decoder()

        im.load()

        self.assertEqual(d.state.xoff, 0)
        self.assertEqual(d.state.yoff, 0)
        self.assertEqual(d.state.xsize, 200)
        self.assertEqual(d.state.ysize, 200)

    def test_negsize(self):
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)
        im.tile = [("MOCK", (xoff, yoff, -10, yoff + ysize), 32, None)]
        self.get_decoder()

        self.assertRaises(ValueError, im.load)

        im.tile = [("MOCK", (xoff, yoff, xoff + xsize, -10), 32, None)]
        self.assertRaises(ValueError, im.load)

    def test_oversize(self):
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)
        im.tile = [("MOCK", (xoff, yoff, xoff + xsize + 100, yoff + ysize), 32, None)]
        self.get_decoder()

        self.assertRaises(ValueError, im.load)

        im.tile = [("MOCK", (xoff, yoff, xoff + xsize, yoff + ysize + 100), 32, None)]
        self.assertRaises(ValueError, im.load)

    def test_no_format(self):
        buf = BytesIO(b"\x00" * 255)

        im = MockImageFile(buf)
        self.assertIsNone(im.format)
        self.assertIsNone(im.get_format_mimetype())

    def test_exif_jpeg(self):
        im = Image.open("Tests/images/exif-72dpi-int.jpg")  # Little endian
        exif = im.getexif()
        self.assertNotIn(258, exif)
        self.assertIn(40960, exif)
        self.assertEqual(exif[40963], 450)
        self.assertEqual(exif[11], "gThumb 3.0.1")

        out = self.tempfile("temp.jpg")
        exif[258] = 8
        del exif[40960]
        exif[40963] = 455
        exif[11] = "Pillow test"
        im.save(out, exif=exif)
        reloaded = Image.open(out)
        reloaded_exif = reloaded.getexif()
        self.assertEqual(reloaded_exif[258], 8)
        self.assertNotIn(40960, exif)
        self.assertEqual(reloaded_exif[40963], 455)
        self.assertEqual(exif[11], "Pillow test")

        im = Image.open("Tests/images/no-dpi-in-exif.jpg")  # Big endian
        exif = im.getexif()
        self.assertNotIn(258, exif)
        self.assertIn(40962, exif)
        self.assertEqual(exif[40963], 200)
        self.assertEqual(exif[305], "Adobe Photoshop CC 2017 (Macintosh)")

        out = self.tempfile("temp.jpg")
        exif[258] = 8
        del exif[34665]
        exif[40963] = 455
        exif[305] = "Pillow test"
        im.save(out, exif=exif)
        reloaded = Image.open(out)
        reloaded_exif = reloaded.getexif()
        self.assertEqual(reloaded_exif[258], 8)
        self.assertNotIn(40960, exif)
        self.assertEqual(reloaded_exif[40963], 455)
        self.assertEqual(exif[305], "Pillow test")

    @unittest.skipIf(
        not HAVE_WEBP or not _webp.HAVE_WEBPANIM,
        "WebP support not installed with animation",
    )
    def test_exif_webp(self):
        im = Image.open("Tests/images/hopper.webp")
        exif = im.getexif()
        self.assertEqual(exif, {})

        out = self.tempfile("temp.webp")
        exif[258] = 8
        exif[40963] = 455
        exif[305] = "Pillow test"

        def check_exif():
            reloaded = Image.open(out)
            reloaded_exif = reloaded.getexif()
            self.assertEqual(reloaded_exif[258], 8)
            self.assertEqual(reloaded_exif[40963], 455)
            self.assertEqual(exif[305], "Pillow test")

        im.save(out, exif=exif)
        check_exif()
        im.save(out, exif=exif, save_all=True)
        check_exif()

    def test_exif_png(self):
        im = Image.open("Tests/images/exif.png")
        exif = im.getexif()
        self.assertEqual(exif, {274: 1})

        out = self.tempfile("temp.png")
        exif[258] = 8
        del exif[274]
        exif[40963] = 455
        exif[305] = "Pillow test"
        im.save(out, exif=exif)

        reloaded = Image.open(out)
        reloaded_exif = reloaded.getexif()
        self.assertEqual(reloaded_exif, {258: 8, 40963: 455, 305: "Pillow test"})

    def test_exif_interop(self):
        im = Image.open("Tests/images/flower.jpg")
        exif = im.getexif()
        self.assertEqual(
            exif.get_ifd(0xA005), {1: "R98", 2: b"0100", 4097: 2272, 4098: 1704}
        )

    def test_exif_shared(self):
        im = Image.open("Tests/images/exif.png")
        exif = im.getexif()
        self.assertIs(im.getexif(), exif)

    def test_exif_str(self):
        im = Image.open("Tests/images/exif.png")
        exif = im.getexif()
        self.assertEqual(str(exif), "{274: 1}")



from .helper import PillowTestCase, fromstring, hopper


class TestImageToBitmap(PillowTestCase):
    def test_sanity(self):

        self.assertRaises(ValueError, lambda: hopper().tobitmap())

        im1 = hopper().convert("1")

        bitmap = im1.tobitmap()

        self.assertIsInstance(bitmap, bytes)
        self.assert_image_equal(im1, fromstring(bitmap))



# Tests potential DOS of IcnsImagePlugin with 0 length block.
# Run from anywhere that PIL is importable.

from io import BytesIO

from PIL import Image
from PIL._util import py3

if py3:
    Image.open(BytesIO(bytes("icns\x00\x00\x00\x10hang\x00\x00\x00\x00", "latin-1")))
else:
    Image.open(BytesIO(bytes("icns\x00\x00\x00\x10hang\x00\x00\x00\x00")))



from PIL import GdImageFile

from .helper import PillowTestCase

TEST_GD_FILE = "Tests/images/hopper.gd"


class TestFileGd(PillowTestCase):
    def test_sanity(self):
        im = GdImageFile.open(TEST_GD_FILE)
        self.assertEqual(im.size, (128, 128))
        self.assertEqual(im.format, "GD")

    def test_bad_mode(self):
        self.assertRaises(ValueError, GdImageFile.open, TEST_GD_FILE, "bad mode")

    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(IOError, GdImageFile.open, invalid_file)



import PIL
import PIL.Image

from .helper import PillowTestCase


class TestSanity(PillowTestCase):
    def test_sanity(self):

        # Make sure we have the binary extension
        PIL.Image.core.new("L", (100, 100))

        # Create an image and do stuff with it.
        im = PIL.Image.new("1", (100, 100))
        self.assertEqual((im.mode, im.size), ("1", (100, 100)))
        self.assertEqual(len(im.tobytes()), 1300)

        # Create images in all remaining major modes.
        PIL.Image.new("L", (100, 100))
        PIL.Image.new("P", (100, 100))
        PIL.Image.new("RGB", (100, 100))
        PIL.Image.new("I", (100, 100))
        PIL.Image.new("F", (100, 100))



import sys
import warnings

from PIL import ImageQt

from .helper import PillowTestCase, hopper

if sys.version_info.major >= 3:
    from importlib import reload

if ImageQt.qt_is_installed:
    from PIL.ImageQt import qRgba

    def skip_if_qt_is_not_installed(_):
        pass


else:

    def skip_if_qt_is_not_installed(test_case):
        test_case.skipTest("Qt bindings are not installed")


class PillowQtTestCase(object):
    def setUp(self):
        skip_if_qt_is_not_installed(self)

    def tearDown(self):
        pass


class PillowQPixmapTestCase(PillowQtTestCase):
    def setUp(self):
        PillowQtTestCase.setUp(self)
        try:
            if ImageQt.qt_version == "5":
                from PyQt5.QtGui import QGuiApplication
            elif ImageQt.qt_version == "4":
                from PyQt4.QtGui import QGuiApplication
            elif ImageQt.qt_version == "side":
                from PySide.QtGui import QGuiApplication
            elif ImageQt.qt_version == "side2":
                from PySide2.QtGui import QGuiApplication
        except ImportError:
            self.skipTest("QGuiApplication not installed")

        self.app = QGuiApplication([])

    def tearDown(self):
        PillowQtTestCase.tearDown(self)
        self.app.quit()


class TestImageQt(PillowQtTestCase, PillowTestCase):
    def test_rgb(self):
        # from https://doc.qt.io/archives/qt-4.8/qcolor.html
        # typedef QRgb
        # An ARGB quadruplet on the format #AARRGGBB,
        # equivalent to an unsigned int.
        if ImageQt.qt_version == "5":
            from PyQt5.QtGui import qRgb
        elif ImageQt.qt_version == "4":
            from PyQt4.QtGui import qRgb
        elif ImageQt.qt_version == "side":
            from PySide.QtGui import qRgb
        elif ImageQt.qt_version == "side2":
            from PySide2.QtGui import qRgb

        self.assertEqual(qRgb(0, 0, 0), qRgba(0, 0, 0, 255))

        def checkrgb(r, g, b):
            val = ImageQt.rgb(r, g, b)
            val = val % 2 ** 24  # drop the alpha
            self.assertEqual(val >> 16, r)
            self.assertEqual(((val >> 8) % 2 ** 8), g)
            self.assertEqual(val % 2 ** 8, b)

        checkrgb(0, 0, 0)
        checkrgb(255, 0, 0)
        checkrgb(0, 255, 0)
        checkrgb(0, 0, 255)

    def test_image(self):
        for mode in ("1", "RGB", "RGBA", "L", "P"):
            ImageQt.ImageQt(hopper(mode))

    def test_deprecated(self):
        with warnings.catch_warnings(record=True) as w:
            reload(ImageQt)
        if ImageQt.qt_version in ["4", "side"]:
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
        else:
            # No warning.
            self.assertEqual(w, [])



import os

from build import vc_setup
from config import all_compilers, bit_from_env, compiler_from_env, compilers, libs
from fetch import fetch
from untar import untar
from unzip import unzip


def _relpath(*args):
    return os.path.join(os.getcwd(), *args)


build_dir = _relpath("build")
inc_dir = _relpath("depends")


def check_sig(filename, signame):
    # UNDONE -- need gpg
    return filename


def mkdirs():
    try:
        os.mkdir(build_dir)
    except OSError:
        pass
    try:
        os.mkdir(inc_dir)
    except OSError:
        pass
    for compiler in all_compilers():
        try:
            os.mkdir(os.path.join(inc_dir, compiler["inc_dir"]))
        except OSError:
            pass


def extract(src, dest):
    if ".zip" in src:
        return unzip(src, dest)
    if ".tar.gz" in src or ".tgz" in src:
        return untar(src, dest)


def extract_libs():
    for name, lib in libs.items():
        filename = lib["filename"]
        if not os.path.exists(filename):
            filename = fetch(lib["url"])
        if name == "openjpeg":
            for compiler in all_compilers():
                if not os.path.exists(
                    os.path.join(build_dir, lib["dir"] + compiler["inc_dir"])
                ):
                    extract(filename, build_dir)
                    os.rename(
                        os.path.join(build_dir, lib["dir"]),
                        os.path.join(build_dir, lib["dir"] + compiler["inc_dir"]),
                    )
        else:
            extract(filename, build_dir)


def extract_openjpeg(compiler):
    return (
        r"""
rem build openjpeg
setlocal
cd %%BUILD%%
mkdir %%INCLIB%%\openjpeg-2.0
copy /Y /B openjpeg-2.0.0-win32-x86\include\openjpeg-2.0  %%INCLIB%%\openjpeg-2.0
copy /Y /B openjpeg-2.0.0-win32-x86\bin\  %%INCLIB%%
copy /Y /B openjpeg-2.0.0-win32-x86\lib\  %%INCLIB%%
endlocal
"""
        % compiler
    )


def cp_tk(ver_85, ver_86):
    versions = {"ver_85": ver_85, "ver_86": ver_86}
    return (
        r"""
mkdir %%INCLIB%%\tcl85\include\X11
copy /Y /B %%BUILD%%\tcl%(ver_85)s\generic\*.h %%INCLIB%%\tcl85\include\
copy /Y /B %%BUILD%%\tk%(ver_85)s\generic\*.h %%INCLIB%%\tcl85\include\
copy /Y /B %%BUILD%%\tk%(ver_85)s\xlib\X11\* %%INCLIB%%\tcl85\include\X11\

mkdir %%INCLIB%%\tcl86\include\X11
copy /Y /B %%BUILD%%\tcl%(ver_86)s\generic\*.h %%INCLIB%%\tcl86\include\
copy /Y /B %%BUILD%%\tk%(ver_86)s\generic\*.h %%INCLIB%%\tcl86\include\
copy /Y /B %%BUILD%%\tk%(ver_86)s\xlib\X11\* %%INCLIB%%\tcl86\include\X11\
"""
        % versions
    )


def header():
    return r"""setlocal
set MSBUILD=C:\Windows\Microsoft.NET\Framework64\v4.0.30319\MSBuild.exe
set CMAKE="cmake.exe"
set INCLIB=%~dp0\depends
set BUILD=%~dp0\build
""" + "\n".join(
        r"set %s=%%BUILD%%\%s" % (k.upper(), v["dir"])
        for (k, v) in libs.items()
        if v["dir"]
    )


def setup_compiler(compiler):
    return (
        r"""setlocal EnableDelayedExpansion
call "%%ProgramFiles%%\Microsoft SDKs\Windows\%(env_version)s\Bin\SetEnv.Cmd" /Release %(env_flags)s
echo on
set INCLIB=%%INCLIB%%\%(inc_dir)s
"""  # noqa: E501
        % compiler
    )


def end_compiler():
    return """
endlocal
"""


def nmake_openjpeg(compiler, bit):
    if compiler["env_version"] == "v7.0":
        return ""

    atts = {"op_ver": "2.3.1"}
    atts.update(compiler)
    return (
        r"""
rem build openjpeg
setlocal
"""
        + vc_setup(compiler, bit)
        + r"""
cd /D %%OPENJPEG%%%(inc_dir)s

%%CMAKE%% -DBUILD_THIRDPARTY:BOOL=OFF -DBUILD_SHARED_LIBS:BOOL=OFF -DCMAKE_BUILD_TYPE=Release -G "NMake Makefiles" .
nmake -nologo -f Makefile clean
nmake -nologo -f Makefile
copy /Y /B bin\* %%INCLIB%%
mkdir %%INCLIB%%\openjpeg-%(op_ver)s
copy /Y /B src\lib\openjp2\*.h %%INCLIB%%\openjpeg-%(op_ver)s
endlocal
"""  # noqa: E501
        % atts
    )


def nmake_libs(compiler, bit):
    # undone -- pre, makes, headers, libs
    script = (
        r"""
rem Build libjpeg
setlocal
"""
        + vc_setup(compiler, bit)
        + r"""
cd /D %%JPEG%%
nmake -nologo -f makefile.vc setup-vc6
nmake -nologo -f makefile.vc clean
nmake -nologo -f makefile.vc nodebug=1 libjpeg.lib
copy /Y /B *.dll %%INCLIB%%
copy /Y /B *.lib %%INCLIB%%
copy /Y /B j*.h %%INCLIB%%
endlocal

rem Build zlib
setlocal
cd /D %%ZLIB%%
nmake -nologo -f win32\Makefile.msc clean
nmake -nologo -f win32\Makefile.msc zlib.lib
copy /Y /B *.dll %%INCLIB%%
copy /Y /B *.lib %%INCLIB%%
copy /Y /B zlib.lib %%INCLIB%%\z.lib
copy /Y /B zlib.h %%INCLIB%%
copy /Y /B zconf.h %%INCLIB%%
endlocal

rem Build webp
setlocal
"""
        + vc_setup(compiler, bit)
        + r"""
cd /D %%WEBP%%
rd /S /Q %%WEBP%%\output\release-static
nmake -nologo -f Makefile.vc CFG=release-static RTLIBCFG=static OBJDIR=output all
copy /Y /B output\release-static\%(webp_platform)s\lib\* %%INCLIB%%
mkdir %%INCLIB%%\webp
copy /Y /B src\webp\*.h %%INCLIB%%\\webp
endlocal

rem Build libtiff
setlocal
"""
        + vc_setup(compiler, bit)
        + r"""
rem do after building jpeg and zlib
copy %%~dp0\nmake.opt %%TIFF%%

cd /D %%TIFF%%
nmake -nologo -f makefile.vc clean
nmake -nologo -f makefile.vc lib
copy /Y /B libtiff\*.dll %%INCLIB%%
copy /Y /B libtiff\*.lib %%INCLIB%%
copy /Y /B libtiff\tiff*.h %%INCLIB%%
endlocal
"""
    )
    return script % compiler


def msbuild_freetype(compiler, bit):
    script = r"""
rem Build freetype
setlocal
rd /S /Q %%FREETYPE%%\objs
set DefaultPlatformToolset=v100
"""
    properties = r"""/p:Configuration="Release" /p:Platform=%(platform)s"""
    if bit == 64:
        script += (
            r"copy /Y /B "
            + r'"C:\Program Files (x86)\Microsoft SDKs\Windows\v7.1A\Lib\x64\*.Lib" '
            + r"%%FREETYPE%%\builds\windows\vc2010"
        )
        properties += r" /p:_IsNativeEnvironment=false"
    script += (
        r"""
%%MSBUILD%% %%FREETYPE%%\builds\windows\vc2010\freetype.sln /t:Clean;Build """
        + properties
        + r""" /m
xcopy /Y /E /Q %%FREETYPE%%\include %%INCLIB%%
"""
    )
    freetypeReleaseDir = r"%%FREETYPE%%\objs\%(platform)s\Release"
    script += (
        r"""
copy /Y /B """
        + freetypeReleaseDir
        + r"""\freetype.lib %%INCLIB%%\freetype.lib
copy /Y /B """
        + freetypeReleaseDir
        + r"""\freetype.dll %%INCLIB%%\..\freetype.dll
endlocal
"""
    )
    return script % compiler


def build_lcms2(compiler):
    if compiler["env_version"] == "v7.1":
        return build_lcms_71(compiler)
    return build_lcms_70(compiler)


def build_lcms_70(compiler):
    """Link error here on x64"""
    if compiler["platform"] == "x64":
        return ""

    """Build LCMS on VC2008. This version is only 32bit/Win32"""
    return (
        r"""
rem Build lcms2
setlocal
rd /S /Q %%LCMS%%\Lib
rd /S /Q %%LCMS%%\Projects\VC%(vc_version)s\Release
%%MSBUILD%% %%LCMS%%\Projects\VC%(vc_version)s\lcms2.sln /t:Clean /p:Configuration="Release" /p:Platform=Win32 /m
%%MSBUILD%% %%LCMS%%\Projects\VC%(vc_version)s\lcms2.sln /t:lcms2_static /p:Configuration="Release" /p:Platform=Win32 /p:PlatformToolset=v90 /m
xcopy /Y /E /Q %%LCMS%%\include %%INCLIB%%
copy /Y /B %%LCMS%%\Lib\MS\*.lib %%INCLIB%%
endlocal
"""  # noqa: E501
        % compiler
    )


def build_lcms_71(compiler):
    return (
        r"""
rem Build lcms2
setlocal
rd /S /Q %%LCMS%%\Lib
rd /S /Q %%LCMS%%\Projects\VC%(vc_version)s\Release
%%MSBUILD%% %%LCMS%%\Projects\VC%(vc_version)s\lcms2.sln /t:Clean /p:Configuration="Release" /p:Platform=%(platform)s /m
%%MSBUILD%% %%LCMS%%\Projects\VC%(vc_version)s\lcms2.sln /t:lcms2_static /p:Configuration="Release" /p:Platform=%(platform)s /m
xcopy /Y /E /Q %%LCMS%%\include %%INCLIB%%
copy /Y /B %%LCMS%%\Lib\MS\*.lib %%INCLIB%%
endlocal
"""  # noqa: E501
        % compiler
    )


def build_ghostscript(compiler, bit):
    script = (
        r"""
rem Build gs
setlocal
"""
        + vc_setup(compiler, bit)
        + r"""
set MSVC_VERSION="""
        + {"2010": "90", "2015": "14"}[compiler["vc_version"]]
        + r"""
set RCOMP="C:\Program Files (x86)\Microsoft SDKs\Windows\v7.1A\Bin\RC.Exe"
cd /D %%GHOSTSCRIPT%%
"""
    )
    if bit == 64:
        script += r"""
set WIN64=""
"""
    script += r"""
nmake -nologo -f psi/msvc.mak
copy /Y /B bin\ C:\Python27\
endlocal
"""
    return script % compiler


def add_compiler(compiler, bit):
    script.append(setup_compiler(compiler))
    script.append(nmake_libs(compiler, bit))

    # script.append(extract_openjpeg(compiler))

    script.append(msbuild_freetype(compiler, bit))
    script.append(build_lcms2(compiler))
    script.append(nmake_openjpeg(compiler, bit))
    script.append(build_ghostscript(compiler, bit))
    script.append(end_compiler())


mkdirs()
extract_libs()
script = [header(), cp_tk(libs["tk-8.5"]["version"], libs["tk-8.6"]["version"])]


if "PYTHON" in os.environ:
    add_compiler(compiler_from_env(), bit_from_env())
else:
    # for compiler in all_compilers():
    #     add_compiler(compiler)
    add_compiler(compilers[7.0][2010][32], 32)

with open("build_deps.cmd", "w") as f:
    f.write("\n".join(script))



from .helper import PillowTestCase, hopper


class TestImageGetPalette(PillowTestCase):
    def test_palette(self):
        def palette(mode):
            p = hopper(mode).getpalette()
            if p:
                return p[:10]
            return None

        self.assertIsNone(palette("1"))
        self.assertIsNone(palette("L"))
        self.assertIsNone(palette("I"))
        self.assertIsNone(palette("F"))
        self.assertEqual(palette("P"), [0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        self.assertIsNone(palette("RGB"))
        self.assertIsNone(palette("RGBA"))
        self.assertIsNone(palette("CMYK"))
        self.assertIsNone(palette("YCbCr"))



from PIL import Image, WmfImagePlugin

from .helper import PillowTestCase, hopper


class TestFileWmf(PillowTestCase):
    def test_load_raw(self):

        # Test basic EMF open and rendering
        im = Image.open("Tests/images/drawing.emf")
        if hasattr(Image.core, "drawwmf"):
            # Currently, support for WMF/EMF is Windows-only
            im.load()
            # Compare to reference rendering
            imref = Image.open("Tests/images/drawing_emf_ref.png")
            imref.load()
            self.assert_image_similar(im, imref, 0)

        # Test basic WMF open and rendering
        im = Image.open("Tests/images/drawing.wmf")
        if hasattr(Image.core, "drawwmf"):
            # Currently, support for WMF/EMF is Windows-only
            im.load()
            # Compare to reference rendering
            imref = Image.open("Tests/images/drawing_wmf_ref.png")
            imref.load()
            self.assert_image_similar(im, imref, 2.0)

    def test_register_handler(self):
        class TestHandler:
            methodCalled = False

            def save(self, im, fp, filename):
                self.methodCalled = True

        handler = TestHandler()
        WmfImagePlugin.register_handler(handler)

        im = hopper()
        tmpfile = self.tempfile("temp.wmf")
        im.save(tmpfile)
        self.assertTrue(handler.methodCalled)

        # Restore the state before this test
        WmfImagePlugin.register_handler(None)

    def test_load_dpi_rounding(self):
        # Round up
        im = Image.open("Tests/images/drawing.emf")
        self.assertEqual(im.info["dpi"], 1424)

        # Round down
        im = Image.open("Tests/images/drawing_roundDown.emf")
        self.assertEqual(im.info["dpi"], 1426)

    def test_save(self):
        im = hopper()

        for ext in [".wmf", ".emf"]:
            tmpfile = self.tempfile("temp" + ext)
            self.assertRaises(IOError, im.save, tmpfile)



from PIL import Image

from .helper import PillowTestCase


class TestFileFtex(PillowTestCase):
    def test_load_raw(self):
        im = Image.open("Tests/images/ftex_uncompressed.ftu")
        target = Image.open("Tests/images/ftex_uncompressed.png")

        self.assert_image_equal(im, target)

    def test_load_dxt1(self):
        im = Image.open("Tests/images/ftex_dxt1.ftc")
        target = Image.open("Tests/images/ftex_dxt1.png")
        self.assert_image_similar(im, target.convert("RGBA"), 15)



from PIL import FitsStubImagePlugin, Image

from .helper import PillowTestCase

TEST_FILE = "Tests/images/hopper.fits"


class TestFileFitsStub(PillowTestCase):
    def test_open(self):
        # Act
        im = Image.open(TEST_FILE)

        # Assert
        self.assertEqual(im.format, "FITS")

        # Dummy data from the stub
        self.assertEqual(im.mode, "F")
        self.assertEqual(im.size, (1, 1))

    def test_invalid_file(self):
        # Arrange
        invalid_file = "Tests/images/flower.jpg"

        # Act / Assert
        self.assertRaises(
            SyntaxError, FitsStubImagePlugin.FITSStubImageFile, invalid_file
        )

    def test_load(self):
        # Arrange
        im = Image.open(TEST_FILE)

        # Act / Assert: stub cannot load without an implemented handler
        self.assertRaises(IOError, im.load)

    def test_save(self):
        # Arrange
        im = Image.open(TEST_FILE)
        dummy_fp = None
        dummy_filename = "dummy.filename"

        # Act / Assert: stub cannot save without an implemented handler
        self.assertRaises(IOError, im.save, dummy_filename)
        self.assertRaises(
            IOError, FitsStubImagePlugin._save, im, dummy_fp, dummy_filename
        )



import io
import struct

from PIL import Image, TiffImagePlugin, TiffTags
from PIL.TiffImagePlugin import IFDRational, _limit_rational

from .helper import PillowTestCase, hopper

tag_ids = {info.name: info.value for info in TiffTags.TAGS_V2.values()}


class TestFileTiffMetadata(PillowTestCase):
    def test_rt_metadata(self):
        """ Test writing arbitrary metadata into the tiff image directory
            Use case is ImageJ private tags, one numeric, one arbitrary
            data.  https://github.com/python-pillow/Pillow/issues/291
            """

        img = hopper()

        # Behaviour change: re #1416
        # Pre ifd rewrite, ImageJMetaData was being written as a string(2),
        # Post ifd rewrite, it's defined as arbitrary bytes(7). It should
        # roundtrip with the actual bytes, rather than stripped text
        # of the premerge tests.
        #
        # For text items, we still have to decode('ascii','replace') because
        # the tiff file format can't take 8 bit bytes in that field.

        basetextdata = "This is some arbitrary metadata for a text field"
        bindata = basetextdata.encode("ascii") + b" \xff"
        textdata = basetextdata + " " + chr(255)
        reloaded_textdata = basetextdata + " ?"
        floatdata = 12.345
        doubledata = 67.89
        info = TiffImagePlugin.ImageFileDirectory()

        ImageJMetaData = tag_ids["ImageJMetaData"]
        ImageJMetaDataByteCounts = tag_ids["ImageJMetaDataByteCounts"]
        ImageDescription = tag_ids["ImageDescription"]

        info[ImageJMetaDataByteCounts] = len(bindata)
        info[ImageJMetaData] = bindata
        info[tag_ids["RollAngle"]] = floatdata
        info.tagtype[tag_ids["RollAngle"]] = 11
        info[tag_ids["YawAngle"]] = doubledata
        info.tagtype[tag_ids["YawAngle"]] = 12

        info[ImageDescription] = textdata

        f = self.tempfile("temp.tif")

        img.save(f, tiffinfo=info)

        loaded = Image.open(f)

        self.assertEqual(loaded.tag[ImageJMetaDataByteCounts], (len(bindata),))
        self.assertEqual(loaded.tag_v2[ImageJMetaDataByteCounts], (len(bindata),))

        self.assertEqual(loaded.tag[ImageJMetaData], bindata)
        self.assertEqual(loaded.tag_v2[ImageJMetaData], bindata)

        self.assertEqual(loaded.tag[ImageDescription], (reloaded_textdata,))
        self.assertEqual(loaded.tag_v2[ImageDescription], reloaded_textdata)

        loaded_float = loaded.tag[tag_ids["RollAngle"]][0]
        self.assertAlmostEqual(loaded_float, floatdata, places=5)
        loaded_double = loaded.tag[tag_ids["YawAngle"]][0]
        self.assertAlmostEqual(loaded_double, doubledata)

        # check with 2 element ImageJMetaDataByteCounts, issue #2006

        info[ImageJMetaDataByteCounts] = (8, len(bindata) - 8)
        img.save(f, tiffinfo=info)
        loaded = Image.open(f)

        self.assertEqual(loaded.tag[ImageJMetaDataByteCounts], (8, len(bindata) - 8))
        self.assertEqual(loaded.tag_v2[ImageJMetaDataByteCounts], (8, len(bindata) - 8))

    def test_read_metadata(self):
        img = Image.open("Tests/images/hopper_g4.tif")

        self.assertEqual(
            {
                "YResolution": IFDRational(4294967295, 113653537),
                "PlanarConfiguration": 1,
                "BitsPerSample": (1,),
                "ImageLength": 128,
                "Compression": 4,
                "FillOrder": 1,
                "RowsPerStrip": 128,
                "ResolutionUnit": 3,
                "PhotometricInterpretation": 0,
                "PageNumber": (0, 1),
                "XResolution": IFDRational(4294967295, 113653537),
                "ImageWidth": 128,
                "Orientation": 1,
                "StripByteCounts": (1968,),
                "SamplesPerPixel": 1,
                "StripOffsets": (8,),
            },
            img.tag_v2.named(),
        )

        self.assertEqual(
            {
                "YResolution": ((4294967295, 113653537),),
                "PlanarConfiguration": (1,),
                "BitsPerSample": (1,),
                "ImageLength": (128,),
                "Compression": (4,),
                "FillOrder": (1,),
                "RowsPerStrip": (128,),
                "ResolutionUnit": (3,),
                "PhotometricInterpretation": (0,),
                "PageNumber": (0, 1),
                "XResolution": ((4294967295, 113653537),),
                "ImageWidth": (128,),
                "Orientation": (1,),
                "StripByteCounts": (1968,),
                "SamplesPerPixel": (1,),
                "StripOffsets": (8,),
            },
            img.tag.named(),
        )

    def test_write_metadata(self):
        """ Test metadata writing through the python code """
        img = Image.open("Tests/images/hopper.tif")

        f = self.tempfile("temp.tiff")
        img.save(f, tiffinfo=img.tag)

        loaded = Image.open(f)

        original = img.tag_v2.named()
        reloaded = loaded.tag_v2.named()

        for k, v in original.items():
            if isinstance(v, IFDRational):
                original[k] = IFDRational(*_limit_rational(v, 2 ** 31))
            elif isinstance(v, tuple) and isinstance(v[0], IFDRational):
                original[k] = tuple(
                    IFDRational(*_limit_rational(elt, 2 ** 31)) for elt in v
                )

        ignored = ["StripByteCounts", "RowsPerStrip", "PageNumber", "StripOffsets"]

        for tag, value in reloaded.items():
            if tag in ignored:
                continue
            if isinstance(original[tag], tuple) and isinstance(
                original[tag][0], IFDRational
            ):
                # Need to compare element by element in the tuple,
                # not comparing tuples of object references
                self.assert_deep_equal(
                    original[tag],
                    value,
                    "%s didn't roundtrip, %s, %s" % (tag, original[tag], value),
                )
            else:
                self.assertEqual(
                    original[tag],
                    value,
                    "%s didn't roundtrip, %s, %s" % (tag, original[tag], value),
                )

        for tag, value in original.items():
            if tag not in ignored:
                self.assertEqual(value, reloaded[tag], "%s didn't roundtrip" % tag)

    def test_no_duplicate_50741_tag(self):
        self.assertEqual(tag_ids["MakerNoteSafety"], 50741)
        self.assertEqual(tag_ids["BestQualityScale"], 50780)

    def test_empty_metadata(self):
        f = io.BytesIO(b"II*\x00\x08\x00\x00\x00")
        head = f.read(8)
        info = TiffImagePlugin.ImageFileDirectory(head)
        # Should not raise struct.error.
        self.assert_warning(UserWarning, info.load, f)

    def test_iccprofile(self):
        # https://github.com/python-pillow/Pillow/issues/1462
        im = Image.open("Tests/images/hopper.iccprofile.tif")
        out = self.tempfile("temp.tiff")

        im.save(out)
        reloaded = Image.open(out)
        self.assertNotIsInstance(im.info["icc_profile"], tuple)
        self.assertEqual(im.info["icc_profile"], reloaded.info["icc_profile"])

    def test_iccprofile_binary(self):
        # https://github.com/python-pillow/Pillow/issues/1526
        # We should be able to load this,
        # but probably won't be able to save it.

        im = Image.open("Tests/images/hopper.iccprofile_binary.tif")
        self.assertEqual(im.tag_v2.tagtype[34675], 1)
        self.assertTrue(im.info["icc_profile"])

    def test_iccprofile_save_png(self):
        im = Image.open("Tests/images/hopper.iccprofile.tif")
        outfile = self.tempfile("temp.png")
        im.save(outfile)

    def test_iccprofile_binary_save_png(self):
        im = Image.open("Tests/images/hopper.iccprofile_binary.tif")
        outfile = self.tempfile("temp.png")
        im.save(outfile)

    def test_exif_div_zero(self):
        im = hopper()
        info = TiffImagePlugin.ImageFileDirectory_v2()
        info[41988] = TiffImagePlugin.IFDRational(0, 0)

        out = self.tempfile("temp.tiff")
        im.save(out, tiffinfo=info, compression="raw")

        reloaded = Image.open(out)
        self.assertEqual(0, reloaded.tag_v2[41988].numerator)
        self.assertEqual(0, reloaded.tag_v2[41988].denominator)

    def test_empty_values(self):
        data = io.BytesIO(
            b"II*\x00\x08\x00\x00\x00\x03\x00\x1a\x01\x05\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x1b\x01\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x98\x82\x02\x00\x07\x00\x00\x002\x00\x00\x00\x00\x00\x00\x00a "
            b"text\x00\x00"
        )
        head = data.read(8)
        info = TiffImagePlugin.ImageFileDirectory_v2(head)
        info.load(data)
        # Should not raise ValueError.
        info = dict(info)
        self.assertIn(33432, info)

    def test_PhotoshopInfo(self):
        im = Image.open("Tests/images/issue_2278.tif")

        self.assertEqual(len(im.tag_v2[34377]), 1)
        self.assertIsInstance(im.tag_v2[34377][0], bytes)
        out = self.tempfile("temp.tiff")
        im.save(out)
        reloaded = Image.open(out)
        self.assertEqual(len(reloaded.tag_v2[34377]), 1)
        self.assertIsInstance(reloaded.tag_v2[34377][0], bytes)

    def test_too_many_entries(self):
        ifd = TiffImagePlugin.ImageFileDirectory_v2()

        #    277: ("SamplesPerPixel", SHORT, 1),
        ifd._tagdata[277] = struct.pack("hh", 4, 4)
        ifd.tagtype[277] = TiffTags.SHORT

        # Should not raise ValueError.
        self.assert_warning(UserWarning, lambda: ifd[277])



from fractions import Fraction

from PIL import Image, TiffImagePlugin
from PIL.TiffImagePlugin import IFDRational

from .helper import PillowTestCase, hopper


class Test_IFDRational(PillowTestCase):
    def _test_equal(self, num, denom, target):

        t = IFDRational(num, denom)

        self.assertEqual(target, t)
        self.assertEqual(t, target)

    def test_sanity(self):

        self._test_equal(1, 1, 1)
        self._test_equal(1, 1, Fraction(1, 1))

        self._test_equal(2, 2, 1)
        self._test_equal(1.0, 1, Fraction(1, 1))

        self._test_equal(Fraction(1, 1), 1, Fraction(1, 1))
        self._test_equal(IFDRational(1, 1), 1, 1)

        self._test_equal(1, 2, Fraction(1, 2))
        self._test_equal(1, 2, IFDRational(1, 2))

    def test_nonetype(self):
        # Fails if the _delegate function doesn't return a valid function

        xres = IFDRational(72)
        yres = IFDRational(72)
        self.assertIsNotNone(xres._val)
        self.assertIsNotNone(xres.numerator)
        self.assertIsNotNone(xres.denominator)
        self.assertIsNotNone(yres._val)

        self.assertTrue(xres and 1)
        self.assertTrue(xres and yres)

    def test_ifd_rational_save(self):
        methods = (True, False)
        if "libtiff_encoder" not in dir(Image.core):
            methods = (False,)

        for libtiff in methods:
            TiffImagePlugin.WRITE_LIBTIFF = libtiff

            im = hopper()
            out = self.tempfile("temp.tiff")
            res = IFDRational(301, 1)
            im.save(out, dpi=(res, res), compression="raw")

            reloaded = Image.open(out)
            self.assertEqual(float(IFDRational(301, 1)), float(reloaded.tag_v2[282]))



from PIL import Image

from .helper import PillowTestCase, hopper


class TestModeI16(PillowTestCase):

    original = hopper().resize((32, 32)).convert("I")

    def verify(self, im1):
        im2 = self.original.copy()
        self.assertEqual(im1.size, im2.size)
        pix1 = im1.load()
        pix2 = im2.load()
        for y in range(im1.size[1]):
            for x in range(im1.size[0]):
                xy = x, y
                p1 = pix1[xy]
                p2 = pix2[xy]
                self.assertEqual(
                    p1,
                    p2,
                    ("got %r from mode %s at %s, expected %r" % (p1, im1.mode, xy, p2)),
                )

    def test_basic(self):
        # PIL 1.1 has limited support for 16-bit image data.  Check that
        # create/copy/transform and save works as expected.

        def basic(mode):

            imIn = self.original.convert(mode)
            self.verify(imIn)

            w, h = imIn.size

            imOut = imIn.copy()
            self.verify(imOut)  # copy

            imOut = imIn.transform((w, h), Image.EXTENT, (0, 0, w, h))
            self.verify(imOut)  # transform

            filename = self.tempfile("temp.im")
            imIn.save(filename)

            imOut = Image.open(filename)

            self.verify(imIn)
            self.verify(imOut)

            imOut = imIn.crop((0, 0, w, h))
            self.verify(imOut)

            imOut = Image.new(mode, (w, h), None)
            imOut.paste(imIn.crop((0, 0, w // 2, h)), (0, 0))
            imOut.paste(imIn.crop((w // 2, 0, w, h)), (w // 2, 0))

            self.verify(imIn)
            self.verify(imOut)

            imIn = Image.new(mode, (1, 1), 1)
            self.assertEqual(imIn.getpixel((0, 0)), 1)

            imIn.putpixel((0, 0), 2)
            self.assertEqual(imIn.getpixel((0, 0)), 2)

            if mode == "L":
                maximum = 255
            else:
                maximum = 32767

            imIn = Image.new(mode, (1, 1), 256)
            self.assertEqual(imIn.getpixel((0, 0)), min(256, maximum))

            imIn.putpixel((0, 0), 512)
            self.assertEqual(imIn.getpixel((0, 0)), min(512, maximum))

        basic("L")

        basic("I;16")
        basic("I;16B")
        basic("I;16L")

        basic("I")

    def test_tobytes(self):
        def tobytes(mode):
            return Image.new(mode, (1, 1), 1).tobytes()

        order = 1 if Image._ENDIAN == "<" else -1

        self.assertEqual(tobytes("L"), b"\x01")
        self.assertEqual(tobytes("I;16"), b"\x01\x00")
        self.assertEqual(tobytes("I;16B"), b"\x00\x01")
        self.assertEqual(tobytes("I"), b"\x01\x00\x00\x00"[::order])

    def test_convert(self):

        im = self.original.copy()

        self.verify(im.convert("I;16"))
        self.verify(im.convert("I;16").convert("L"))
        self.verify(im.convert("I;16").convert("I"))

        self.verify(im.convert("I;16B"))
        self.verify(im.convert("I;16B").convert("L"))
        self.verify(im.convert("I;16B").convert("I"))



from PIL import Image

from .helper import PillowTestCase


class TestPickle(PillowTestCase):
    def helper_pickle_file(self, pickle, protocol=0, mode=None):
        # Arrange
        im = Image.open("Tests/images/hopper.jpg")
        filename = self.tempfile("temp.pkl")
        if mode:
            im = im.convert(mode)

        # Act
        with open(filename, "wb") as f:
            pickle.dump(im, f, protocol)
        with open(filename, "rb") as f:
            loaded_im = pickle.load(f)

        # Assert
        self.assertEqual(im, loaded_im)

    def helper_pickle_string(
        self, pickle, protocol=0, test_file="Tests/images/hopper.jpg", mode=None
    ):
        im = Image.open(test_file)
        if mode:
            im = im.convert(mode)

        # Act
        dumped_string = pickle.dumps(im, protocol)
        loaded_im = pickle.loads(dumped_string)

        # Assert
        self.assertEqual(im, loaded_im)

    def test_pickle_image(self):
        # Arrange
        import pickle

        # Act / Assert
        for protocol in range(0, pickle.HIGHEST_PROTOCOL + 1):
            self.helper_pickle_string(pickle, protocol)
            self.helper_pickle_file(pickle, protocol)

    def test_cpickle_image(self):
        # Arrange
        try:
            import cPickle
        except ImportError:
            return

        # Act / Assert
        for protocol in range(0, cPickle.HIGHEST_PROTOCOL + 1):
            self.helper_pickle_string(cPickle, protocol)
            self.helper_pickle_file(cPickle, protocol)

    def test_pickle_p_mode(self):
        # Arrange
        import pickle

        # Act / Assert
        for test_file in [
            "Tests/images/test-card.png",
            "Tests/images/zero_bb.png",
            "Tests/images/zero_bb_scale2.png",
            "Tests/images/non_zero_bb.png",
            "Tests/images/non_zero_bb_scale2.png",
            "Tests/images/p_trns_single.png",
            "Tests/images/pil123p.png",
            "Tests/images/itxt_chunks.png",
        ]:
            for protocol in range(0, pickle.HIGHEST_PROTOCOL + 1):
                self.helper_pickle_string(
                    pickle, protocol=protocol, test_file=test_file
                )

    def test_pickle_pa_mode(self):
        # Arrange
        import pickle

        # Act / Assert
        for protocol in range(0, pickle.HIGHEST_PROTOCOL + 1):
            self.helper_pickle_string(pickle, protocol, mode="PA")
            self.helper_pickle_file(pickle, protocol, mode="PA")

    def test_pickle_l_mode(self):
        # Arrange
        import pickle

        # Act / Assert
        for protocol in range(0, pickle.HIGHEST_PROTOCOL + 1):
            self.helper_pickle_string(pickle, protocol, mode="L")
            self.helper_pickle_file(pickle, protocol, mode="L")

    def test_pickle_la_mode_with_palette(self):
        # Arrange
        import pickle

        im = Image.open("Tests/images/hopper.jpg")
        filename = self.tempfile("temp.pkl")
        im = im.convert("PA")

        # Act / Assert
        for protocol in range(0, pickle.HIGHEST_PROTOCOL + 1):
            im.mode = "LA"
            with open(filename, "wb") as f:
                pickle.dump(im, f, protocol)
            with open(filename, "rb") as f:
                loaded_im = pickle.load(f)

            im.mode = "PA"
            self.assertEqual(im, loaded_im)

    def test_cpickle_l_mode(self):
        # Arrange
        try:
            import cPickle
        except ImportError:
            return

        # Act / Assert
        for protocol in range(0, cPickle.HIGHEST_PROTOCOL + 1):
            self.helper_pickle_string(cPickle, protocol, mode="L")
            self.helper_pickle_file(cPickle, protocol, mode="L")



import os.path

from PIL import Image, ImageColor, ImageDraw, ImageFont, features

from .helper import PillowTestCase, hopper, unittest

BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
GRAY = (190, 190, 190)
DEFAULT_MODE = "RGB"
IMAGES_PATH = os.path.join("Tests", "images", "imagedraw")

# Image size
W, H = 100, 100

# Bounding box points
X0 = int(W / 4)
X1 = int(X0 * 3)
Y0 = int(H / 4)
Y1 = int(X0 * 3)

# Two kinds of bounding box
BBOX1 = [(X0, Y0), (X1, Y1)]
BBOX2 = [X0, Y0, X1, Y1]

# Two kinds of coordinate sequences
POINTS1 = [(10, 10), (20, 40), (30, 30)]
POINTS2 = [10, 10, 20, 40, 30, 30]

KITE_POINTS = [(10, 50), (70, 10), (90, 50), (70, 90), (10, 50)]

HAS_FREETYPE = features.check("freetype2")


class TestImageDraw(PillowTestCase):
    def test_sanity(self):
        im = hopper("RGB").copy()

        draw = ImageDraw.ImageDraw(im)
        draw = ImageDraw.Draw(im)

        draw.ellipse(list(range(4)))
        draw.line(list(range(10)))
        draw.polygon(list(range(100)))
        draw.rectangle(list(range(4)))

    def test_valueerror(self):
        im = Image.open("Tests/images/chi.gif")

        draw = ImageDraw.Draw(im)
        draw.line((0, 0), fill=(0, 0, 0))

    def test_mode_mismatch(self):
        im = hopper("RGB").copy()

        self.assertRaises(ValueError, ImageDraw.ImageDraw, im, mode="L")

    def helper_arc(self, bbox, start, end):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)

        # Act
        draw.arc(bbox, start, end)

        # Assert
        self.assert_image_similar(im, Image.open("Tests/images/imagedraw_arc.png"), 1)

    def test_arc1(self):
        self.helper_arc(BBOX1, 0, 180)
        self.helper_arc(BBOX1, 0.5, 180.4)

    def test_arc2(self):
        self.helper_arc(BBOX2, 0, 180)
        self.helper_arc(BBOX2, 0.5, 180.4)

    def test_arc_end_le_start(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        start = 270.5
        end = 0

        # Act
        draw.arc(BBOX1, start=start, end=end)

        # Assert
        self.assert_image_equal(
            im, Image.open("Tests/images/imagedraw_arc_end_le_start.png")
        )

    def test_arc_no_loops(self):
        # No need to go in loops
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        start = 5
        end = 370

        # Act
        draw.arc(BBOX1, start=start, end=end)

        # Assert
        self.assert_image_similar(
            im, Image.open("Tests/images/imagedraw_arc_no_loops.png"), 1
        )

    def test_arc_width(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_arc_width.png"

        # Act
        draw.arc(BBOX1, 10, 260, width=5)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_arc_width_pieslice_large(self):
        # Tests an arc with a large enough width that it is a pieslice
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_arc_width_pieslice.png"

        # Act
        draw.arc(BBOX1, 10, 260, fill="yellow", width=100)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_arc_width_fill(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_arc_width_fill.png"

        # Act
        draw.arc(BBOX1, 10, 260, fill="yellow", width=5)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_arc_width_non_whole_angle(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_arc_width_non_whole_angle.png"

        # Act
        draw.arc(BBOX1, 10, 259.5, width=5)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_bitmap(self):
        # Arrange
        small = Image.open("Tests/images/pil123rgba.png").resize((50, 50))
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)

        # Act
        draw.bitmap((10, 10), small)

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_bitmap.png"))

    def helper_chord(self, mode, bbox, start, end):
        # Arrange
        im = Image.new(mode, (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_chord_{}.png".format(mode)

        # Act
        draw.chord(bbox, start, end, fill="red", outline="yellow")

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_chord1(self):
        for mode in ["RGB", "L"]:
            self.helper_chord(mode, BBOX1, 0, 180)
            self.helper_chord(mode, BBOX1, 0.5, 180.4)

    def test_chord2(self):
        for mode in ["RGB", "L"]:
            self.helper_chord(mode, BBOX2, 0, 180)
            self.helper_chord(mode, BBOX2, 0.5, 180.4)

    def test_chord_width(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_chord_width.png"

        # Act
        draw.chord(BBOX1, 10, 260, outline="yellow", width=5)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_chord_width_fill(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_chord_width_fill.png"

        # Act
        draw.chord(BBOX1, 10, 260, fill="red", outline="yellow", width=5)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def helper_ellipse(self, mode, bbox):
        # Arrange
        im = Image.new(mode, (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_ellipse_{}.png".format(mode)

        # Act
        draw.ellipse(bbox, fill="green", outline="blue")

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_ellipse1(self):
        for mode in ["RGB", "L"]:
            self.helper_ellipse(mode, BBOX1)

    def test_ellipse2(self):
        for mode in ["RGB", "L"]:
            self.helper_ellipse(mode, BBOX2)

    def test_ellipse_edge(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)

        # Act
        draw.ellipse(((0, 0), (W - 1, H)), fill="white")

        # Assert
        self.assert_image_similar(
            im, Image.open("Tests/images/imagedraw_ellipse_edge.png"), 1
        )

    def test_ellipse_symmetric(self):
        for bbox in [(25, 25, 76, 76), (25, 25, 75, 75)]:
            im = Image.new("RGB", (101, 101))
            draw = ImageDraw.Draw(im)
            draw.ellipse(bbox, fill="green", outline="blue")
            self.assert_image_equal(im, im.transpose(Image.FLIP_LEFT_RIGHT))

    def test_ellipse_width(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_ellipse_width.png"

        # Act
        draw.ellipse(BBOX1, outline="blue", width=5)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_ellipse_width_large(self):
        # Arrange
        im = Image.new("RGB", (500, 500))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_ellipse_width_large.png"

        # Act
        draw.ellipse((25, 25, 475, 475), outline="blue", width=75)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_ellipse_width_fill(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_ellipse_width_fill.png"

        # Act
        draw.ellipse(BBOX1, fill="green", outline="blue", width=5)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def helper_line(self, points):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)

        # Act
        draw.line(points, fill="yellow", width=2)

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_line.png"))

    def test_line1(self):
        self.helper_line(POINTS1)

    def test_line2(self):
        self.helper_line(POINTS2)

    def test_shape1(self):
        # Arrange
        im = Image.new("RGB", (100, 100), "white")
        draw = ImageDraw.Draw(im)
        x0, y0 = 5, 5
        x1, y1 = 5, 50
        x2, y2 = 95, 50
        x3, y3 = 95, 5

        # Act
        s = ImageDraw.Outline()
        s.move(x0, y0)
        s.curve(x1, y1, x2, y2, x3, y3)
        s.line(x0, y0)

        draw.shape(s, fill=1)

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_shape1.png"))

    def test_shape2(self):
        # Arrange
        im = Image.new("RGB", (100, 100), "white")
        draw = ImageDraw.Draw(im)
        x0, y0 = 95, 95
        x1, y1 = 95, 50
        x2, y2 = 5, 50
        x3, y3 = 5, 95

        # Act
        s = ImageDraw.Outline()
        s.move(x0, y0)
        s.curve(x1, y1, x2, y2, x3, y3)
        s.line(x0, y0)

        draw.shape(s, outline="blue")

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_shape2.png"))

    def helper_pieslice(self, bbox, start, end):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)

        # Act
        draw.pieslice(bbox, start, end, fill="white", outline="blue")

        # Assert
        self.assert_image_similar(
            im, Image.open("Tests/images/imagedraw_pieslice.png"), 1
        )

    def test_pieslice1(self):
        self.helper_pieslice(BBOX1, -90, 45)
        self.helper_pieslice(BBOX1, -90.5, 45.4)

    def test_pieslice2(self):
        self.helper_pieslice(BBOX2, -90, 45)
        self.helper_pieslice(BBOX2, -90.5, 45.4)

    def test_pieslice_width(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_pieslice_width.png"

        # Act
        draw.pieslice(BBOX1, 10, 260, outline="blue", width=5)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_pieslice_width_fill(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_pieslice_width_fill.png"

        # Act
        draw.pieslice(BBOX1, 10, 260, fill="white", outline="blue", width=5)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def helper_point(self, points):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)

        # Act
        draw.point(points, fill="yellow")

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_point.png"))

    def test_point1(self):
        self.helper_point(POINTS1)

    def test_point2(self):
        self.helper_point(POINTS2)

    def helper_polygon(self, points):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)

        # Act
        draw.polygon(points, fill="red", outline="blue")

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_polygon.png"))

    def test_polygon1(self):
        self.helper_polygon(POINTS1)

    def test_polygon2(self):
        self.helper_polygon(POINTS2)

    def test_polygon_kite(self):
        # Test drawing lines of different gradients (dx>dy, dy>dx) and
        # vertical (dx==0) and horizontal (dy==0) lines
        for mode in ["RGB", "L"]:
            # Arrange
            im = Image.new(mode, (W, H))
            draw = ImageDraw.Draw(im)
            expected = "Tests/images/imagedraw_polygon_kite_{}.png".format(mode)

            # Act
            draw.polygon(KITE_POINTS, fill="blue", outline="yellow")

            # Assert
            self.assert_image_equal(im, Image.open(expected))

    def helper_rectangle(self, bbox):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)

        # Act
        draw.rectangle(bbox, fill="black", outline="green")

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_rectangle.png"))

    def test_rectangle1(self):
        self.helper_rectangle(BBOX1)

    def test_rectangle2(self):
        self.helper_rectangle(BBOX2)

    def test_big_rectangle(self):
        # Test drawing a rectangle bigger than the image
        # Arrange
        im = Image.new("RGB", (W, H))
        bbox = [(-1, -1), (W + 1, H + 1)]
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_big_rectangle.png"

        # Act
        draw.rectangle(bbox, fill="orange")

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_rectangle_width(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_rectangle_width.png"

        # Act
        draw.rectangle(BBOX1, outline="green", width=5)

        # Assert
        self.assert_image_equal(im, Image.open(expected))

    def test_rectangle_width_fill(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_rectangle_width_fill.png"

        # Act
        draw.rectangle(BBOX1, fill="blue", outline="green", width=5)

        # Assert
        self.assert_image_equal(im, Image.open(expected))

    def test_rectangle_I16(self):
        # Arrange
        im = Image.new("I;16", (W, H))
        draw = ImageDraw.Draw(im)

        # Act
        draw.rectangle(BBOX1, fill="black", outline="green")

        # Assert
        self.assert_image_equal(
            im.convert("I"), Image.open("Tests/images/imagedraw_rectangle_I.png")
        )

    def test_floodfill(self):
        red = ImageColor.getrgb("red")

        for mode, value in [("L", 1), ("RGBA", (255, 0, 0, 0)), ("RGB", red)]:
            # Arrange
            im = Image.new(mode, (W, H))
            draw = ImageDraw.Draw(im)
            draw.rectangle(BBOX2, outline="yellow", fill="green")
            centre_point = (int(W / 2), int(H / 2))

            # Act
            ImageDraw.floodfill(im, centre_point, value)

            # Assert
            expected = "Tests/images/imagedraw_floodfill_" + mode + ".png"
            im_floodfill = Image.open(expected)
            self.assert_image_equal(im, im_floodfill)

        # Test that using the same colour does not change the image
        ImageDraw.floodfill(im, centre_point, red)
        self.assert_image_equal(im, im_floodfill)

        # Test that filling outside the image does not change the image
        ImageDraw.floodfill(im, (W, H), red)
        self.assert_image_equal(im, im_floodfill)

        # Test filling at the edge of an image
        im = Image.new("RGB", (1, 1))
        ImageDraw.floodfill(im, (0, 0), red)
        self.assert_image_equal(im, Image.new("RGB", (1, 1), red))

    def test_floodfill_border(self):
        # floodfill() is experimental

        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        draw.rectangle(BBOX2, outline="yellow", fill="green")
        centre_point = (int(W / 2), int(H / 2))

        # Act
        ImageDraw.floodfill(
            im,
            centre_point,
            ImageColor.getrgb("red"),
            border=ImageColor.getrgb("black"),
        )

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_floodfill2.png"))

    def test_floodfill_thresh(self):
        # floodfill() is experimental

        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        draw.rectangle(BBOX2, outline="darkgreen", fill="green")
        centre_point = (int(W / 2), int(H / 2))

        # Act
        ImageDraw.floodfill(im, centre_point, ImageColor.getrgb("red"), thresh=30)

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_floodfill2.png"))

    def test_floodfill_not_negative(self):
        # floodfill() is experimental
        # Test that floodfill does not extend into negative coordinates

        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        draw.line((W / 2, 0, W / 2, H / 2), fill="green")
        draw.line((0, H / 2, W / 2, H / 2), fill="green")

        # Act
        ImageDraw.floodfill(im, (int(W / 4), int(H / 4)), ImageColor.getrgb("red"))

        # Assert
        self.assert_image_equal(
            im, Image.open("Tests/images/imagedraw_floodfill_not_negative.png")
        )

    def create_base_image_draw(
        self, size, mode=DEFAULT_MODE, background1=WHITE, background2=GRAY
    ):
        img = Image.new(mode, size, background1)
        for x in range(0, size[0]):
            for y in range(0, size[1]):
                if (x + y) % 2 == 0:
                    img.putpixel((x, y), background2)
        return img, ImageDraw.Draw(img)

    def test_square(self):
        expected = Image.open(os.path.join(IMAGES_PATH, "square.png"))
        expected.load()
        img, draw = self.create_base_image_draw((10, 10))
        draw.polygon([(2, 2), (2, 7), (7, 7), (7, 2)], BLACK)
        self.assert_image_equal(img, expected, "square as normal polygon failed")
        img, draw = self.create_base_image_draw((10, 10))
        draw.polygon([(7, 7), (7, 2), (2, 2), (2, 7)], BLACK)
        self.assert_image_equal(img, expected, "square as inverted polygon failed")
        img, draw = self.create_base_image_draw((10, 10))
        draw.rectangle((2, 2, 7, 7), BLACK)
        self.assert_image_equal(img, expected, "square as normal rectangle failed")
        img, draw = self.create_base_image_draw((10, 10))
        draw.rectangle((7, 7, 2, 2), BLACK)
        self.assert_image_equal(img, expected, "square as inverted rectangle failed")

    def test_triangle_right(self):
        expected = Image.open(os.path.join(IMAGES_PATH, "triangle_right.png"))
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.polygon([(3, 5), (17, 5), (10, 12)], BLACK)
        self.assert_image_equal(img, expected, "triangle right failed")

    def test_line_horizontal(self):
        expected = Image.open(
            os.path.join(IMAGES_PATH, "line_horizontal_w2px_normal.png")
        )
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((5, 5, 14, 5), BLACK, 2)
        self.assert_image_equal(
            img, expected, "line straight horizontal normal 2px wide failed"
        )
        expected = Image.open(
            os.path.join(IMAGES_PATH, "line_horizontal_w2px_inverted.png")
        )
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((14, 5, 5, 5), BLACK, 2)
        self.assert_image_equal(
            img, expected, "line straight horizontal inverted 2px wide failed"
        )
        expected = Image.open(os.path.join(IMAGES_PATH, "line_horizontal_w3px.png"))
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((5, 5, 14, 5), BLACK, 3)
        self.assert_image_equal(
            img, expected, "line straight horizontal normal 3px wide failed"
        )
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((14, 5, 5, 5), BLACK, 3)
        self.assert_image_equal(
            img, expected, "line straight horizontal inverted 3px wide failed"
        )
        expected = Image.open(os.path.join(IMAGES_PATH, "line_horizontal_w101px.png"))
        expected.load()
        img, draw = self.create_base_image_draw((200, 110))
        draw.line((5, 55, 195, 55), BLACK, 101)
        self.assert_image_equal(
            img, expected, "line straight horizontal 101px wide failed"
        )

    def test_line_h_s1_w2(self):
        self.skipTest("failing")
        expected = Image.open(
            os.path.join(IMAGES_PATH, "line_horizontal_slope1px_w2px.png")
        )
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((5, 5, 14, 6), BLACK, 2)
        self.assert_image_equal(
            img, expected, "line horizontal 1px slope 2px wide failed"
        )

    def test_line_vertical(self):
        expected = Image.open(
            os.path.join(IMAGES_PATH, "line_vertical_w2px_normal.png")
        )
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((5, 5, 5, 14), BLACK, 2)
        self.assert_image_equal(
            img, expected, "line straight vertical normal 2px wide failed"
        )
        expected = Image.open(
            os.path.join(IMAGES_PATH, "line_vertical_w2px_inverted.png")
        )
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((5, 14, 5, 5), BLACK, 2)
        self.assert_image_equal(
            img, expected, "line straight vertical inverted 2px wide failed"
        )
        expected = Image.open(os.path.join(IMAGES_PATH, "line_vertical_w3px.png"))
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((5, 5, 5, 14), BLACK, 3)
        self.assert_image_equal(
            img, expected, "line straight vertical normal 3px wide failed"
        )
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((5, 14, 5, 5), BLACK, 3)
        self.assert_image_equal(
            img, expected, "line straight vertical inverted 3px wide failed"
        )
        expected = Image.open(os.path.join(IMAGES_PATH, "line_vertical_w101px.png"))
        expected.load()
        img, draw = self.create_base_image_draw((110, 200))
        draw.line((55, 5, 55, 195), BLACK, 101)
        self.assert_image_equal(
            img, expected, "line straight vertical 101px wide failed"
        )
        expected = Image.open(
            os.path.join(IMAGES_PATH, "line_vertical_slope1px_w2px.png")
        )
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((5, 5, 6, 14), BLACK, 2)
        self.assert_image_equal(
            img, expected, "line vertical 1px slope 2px wide failed"
        )

    def test_line_oblique_45(self):
        expected = Image.open(os.path.join(IMAGES_PATH, "line_oblique_45_w3px_a.png"))
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((5, 5, 14, 14), BLACK, 3)
        self.assert_image_equal(
            img, expected, "line oblique 45 normal 3px wide A failed"
        )
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((14, 14, 5, 5), BLACK, 3)
        self.assert_image_equal(
            img, expected, "line oblique 45 inverted 3px wide A failed"
        )
        expected = Image.open(os.path.join(IMAGES_PATH, "line_oblique_45_w3px_b.png"))
        expected.load()
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((14, 5, 5, 14), BLACK, 3)
        self.assert_image_equal(
            img, expected, "line oblique 45 normal 3px wide B failed"
        )
        img, draw = self.create_base_image_draw((20, 20))
        draw.line((5, 14, 14, 5), BLACK, 3)
        self.assert_image_equal(
            img, expected, "line oblique 45 inverted 3px wide B failed"
        )

    def test_wide_line_dot(self):
        # Test drawing a wide "line" from one point to another just draws
        # a single point
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_wide_line_dot.png"

        # Act
        draw.line([(50, 50), (50, 50)], width=3)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_line_joint(self):
        im = Image.new("RGB", (500, 325))
        draw = ImageDraw.Draw(im)
        expected = "Tests/images/imagedraw_line_joint_curve.png"

        # Act
        xy = [
            (400, 280),
            (380, 280),
            (450, 280),
            (440, 120),
            (350, 200),
            (310, 280),
            (300, 280),
            (250, 280),
            (250, 200),
            (150, 200),
            (150, 260),
            (50, 200),
            (150, 50),
            (250, 100),
        ]
        draw.line(xy, GRAY, 50, "curve")

        # Assert
        self.assert_image_similar(im, Image.open(expected), 3)

    def test_textsize_empty_string(self):
        # https://github.com/python-pillow/Pillow/issues/2783
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)

        # Act
        # Should not cause 'SystemError: <built-in method getsize of
        # ImagingFont object at 0x...> returned NULL without setting an error'
        draw.textsize("")
        draw.textsize("\n")
        draw.textsize("test\n")

    @unittest.skipUnless(HAS_FREETYPE, "ImageFont not available")
    def test_textsize_stroke(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw.Draw(im)
        font = ImageFont.truetype("Tests/fonts/FreeMono.ttf", 20)

        # Act / Assert
        self.assertEqual(draw.textsize("A", font, stroke_width=2), (16, 20))
        self.assertEqual(
            draw.multiline_textsize("ABC\nAaaa", font, stroke_width=2), (52, 44)
        )

    @unittest.skipUnless(HAS_FREETYPE, "ImageFont not available")
    def test_stroke(self):
        for suffix, stroke_fill in {"same": None, "different": "#0f0"}.items():
            # Arrange
            im = Image.new("RGB", (120, 130))
            draw = ImageDraw.Draw(im)
            font = ImageFont.truetype("Tests/fonts/FreeMono.ttf", 120)

            # Act
            draw.text(
                (10, 10), "A", "#f00", font, stroke_width=2, stroke_fill=stroke_fill
            )

            # Assert
            self.assert_image_similar(
                im, Image.open("Tests/images/imagedraw_stroke_" + suffix + ".png"), 3.1
            )

    @unittest.skipUnless(HAS_FREETYPE, "ImageFont not available")
    def test_stroke_multiline(self):
        # Arrange
        im = Image.new("RGB", (100, 250))
        draw = ImageDraw.Draw(im)
        font = ImageFont.truetype("Tests/fonts/FreeMono.ttf", 120)

        # Act
        draw.multiline_text(
            (10, 10), "A\nB", "#f00", font, stroke_width=2, stroke_fill="#0f0"
        )

        # Assert
        self.assert_image_similar(
            im, Image.open("Tests/images/imagedraw_stroke_multiline.png"), 3.3
        )

    def test_same_color_outline(self):
        # Prepare shape
        x0, y0 = 5, 5
        x1, y1 = 5, 50
        x2, y2 = 95, 50
        x3, y3 = 95, 5

        s = ImageDraw.Outline()
        s.move(x0, y0)
        s.curve(x1, y1, x2, y2, x3, y3)
        s.line(x0, y0)

        # Begin
        for mode in ["RGB", "L"]:
            for fill, outline in [["red", None], ["red", "red"], ["red", "#f00"]]:
                for operation, args in {
                    "chord": [BBOX1, 0, 180],
                    "ellipse": [BBOX1],
                    "shape": [s],
                    "pieslice": [BBOX1, -90, 45],
                    "polygon": [[(18, 30), (85, 30), (60, 72)]],
                    "rectangle": [BBOX1],
                }.items():
                    # Arrange
                    im = Image.new(mode, (W, H))
                    draw = ImageDraw.Draw(im)

                    # Act
                    draw_method = getattr(draw, operation)
                    args += [fill, outline]
                    draw_method(*args)

                    # Assert
                    expected = "Tests/images/imagedraw_outline_{}_{}.png".format(
                        operation, mode
                    )
                    self.assert_image_similar(im, Image.open(expected), 1)



import os
import sys
from io import BytesIO

from PIL import Image, ImageFile, JpegImagePlugin

from .helper import PillowTestCase, cjpeg_available, djpeg_available, hopper, unittest

codecs = dir(Image.core)

TEST_FILE = "Tests/images/hopper.jpg"


class TestFileJpeg(PillowTestCase):
    def setUp(self):
        if "jpeg_encoder" not in codecs or "jpeg_decoder" not in codecs:
            self.skipTest("jpeg support not available")

    def roundtrip(self, im, **options):
        out = BytesIO()
        im.save(out, "JPEG", **options)
        test_bytes = out.tell()
        out.seek(0)
        im = Image.open(out)
        im.bytes = test_bytes  # for testing only
        return im

    def gen_random_image(self, size, mode="RGB"):
        """ Generates a very hard to compress file
        :param size: tuple
        :param mode: optional image mode

        """
        return Image.frombytes(mode, size, os.urandom(size[0] * size[1] * len(mode)))

    def test_sanity(self):

        # internal version number
        self.assertRegex(Image.core.jpeglib_version, r"\d+\.\d+$")

        im = Image.open(TEST_FILE)
        im.load()
        self.assertEqual(im.mode, "RGB")
        self.assertEqual(im.size, (128, 128))
        self.assertEqual(im.format, "JPEG")
        self.assertEqual(im.get_format_mimetype(), "image/jpeg")

    def test_app(self):
        # Test APP/COM reader (@PIL135)
        im = Image.open(TEST_FILE)
        self.assertEqual(
            im.applist[0], ("APP0", b"JFIF\x00\x01\x01\x01\x00`\x00`\x00\x00")
        )
        self.assertEqual(
            im.applist[1], ("COM", b"File written by Adobe Photoshop\xa8 4.0\x00")
        )
        self.assertEqual(len(im.applist), 2)

    def test_cmyk(self):
        # Test CMYK handling.  Thanks to Tim and Charlie for test data,
        # Michael for getting me to look one more time.
        f = "Tests/images/pil_sample_cmyk.jpg"
        im = Image.open(f)
        # the source image has red pixels in the upper left corner.
        c, m, y, k = [x / 255.0 for x in im.getpixel((0, 0))]
        self.assertEqual(c, 0.0)
        self.assertGreater(m, 0.8)
        self.assertGreater(y, 0.8)
        self.assertEqual(k, 0.0)
        # the opposite corner is black
        c, m, y, k = [x / 255.0 for x in im.getpixel((im.size[0] - 1, im.size[1] - 1))]
        self.assertGreater(k, 0.9)
        # roundtrip, and check again
        im = self.roundtrip(im)
        c, m, y, k = [x / 255.0 for x in im.getpixel((0, 0))]
        self.assertEqual(c, 0.0)
        self.assertGreater(m, 0.8)
        self.assertGreater(y, 0.8)
        self.assertEqual(k, 0.0)
        c, m, y, k = [x / 255.0 for x in im.getpixel((im.size[0] - 1, im.size[1] - 1))]
        self.assertGreater(k, 0.9)

    def test_dpi(self):
        def test(xdpi, ydpi=None):
            im = Image.open(TEST_FILE)
            im = self.roundtrip(im, dpi=(xdpi, ydpi or xdpi))
            return im.info.get("dpi")

        self.assertEqual(test(72), (72, 72))
        self.assertEqual(test(300), (300, 300))
        self.assertEqual(test(100, 200), (100, 200))
        self.assertIsNone(test(0))  # square pixels

    def test_icc(self):
        # Test ICC support
        im1 = Image.open("Tests/images/rgb.jpg")
        icc_profile = im1.info["icc_profile"]
        self.assertEqual(len(icc_profile), 3144)
        # Roundtrip via physical file.
        f = self.tempfile("temp.jpg")
        im1.save(f, icc_profile=icc_profile)
        im2 = Image.open(f)
        self.assertEqual(im2.info.get("icc_profile"), icc_profile)
        # Roundtrip via memory buffer.
        im1 = self.roundtrip(hopper())
        im2 = self.roundtrip(hopper(), icc_profile=icc_profile)
        self.assert_image_equal(im1, im2)
        self.assertFalse(im1.info.get("icc_profile"))
        self.assertTrue(im2.info.get("icc_profile"))

    def test_icc_big(self):
        # Make sure that the "extra" support handles large blocks
        def test(n):
            # The ICC APP marker can store 65519 bytes per marker, so
            # using a 4-byte test code should allow us to detect out of
            # order issues.
            icc_profile = (b"Test" * int(n / 4 + 1))[:n]
            self.assertEqual(len(icc_profile), n)  # sanity
            im1 = self.roundtrip(hopper(), icc_profile=icc_profile)
            self.assertEqual(im1.info.get("icc_profile"), icc_profile or None)

        test(0)
        test(1)
        test(3)
        test(4)
        test(5)
        test(65533 - 14)  # full JPEG marker block
        test(65533 - 14 + 1)  # full block plus one byte
        test(ImageFile.MAXBLOCK)  # full buffer block
        test(ImageFile.MAXBLOCK + 1)  # full buffer block plus one byte
        test(ImageFile.MAXBLOCK * 4 + 3)  # large block

    def test_large_icc_meta(self):
        # https://github.com/python-pillow/Pillow/issues/148
        # Sometimes the meta data on the icc_profile block is bigger than
        # Image.MAXBLOCK or the image size.
        im = Image.open("Tests/images/icc_profile_big.jpg")
        f = self.tempfile("temp.jpg")
        icc_profile = im.info["icc_profile"]
        # Should not raise IOError for image with icc larger than image size.
        im.save(
            f,
            format="JPEG",
            progressive=True,
            quality=95,
            icc_profile=icc_profile,
            optimize=True,
        )

    def test_optimize(self):
        im1 = self.roundtrip(hopper())
        im2 = self.roundtrip(hopper(), optimize=0)
        im3 = self.roundtrip(hopper(), optimize=1)
        self.assert_image_equal(im1, im2)
        self.assert_image_equal(im1, im3)
        self.assertGreaterEqual(im1.bytes, im2.bytes)
        self.assertGreaterEqual(im1.bytes, im3.bytes)

    def test_optimize_large_buffer(self):
        # https://github.com/python-pillow/Pillow/issues/148
        f = self.tempfile("temp.jpg")
        # this requires ~ 1.5x Image.MAXBLOCK
        im = Image.new("RGB", (4096, 4096), 0xFF3333)
        im.save(f, format="JPEG", optimize=True)

    def test_progressive(self):
        im1 = self.roundtrip(hopper())
        im2 = self.roundtrip(hopper(), progressive=False)
        im3 = self.roundtrip(hopper(), progressive=True)
        self.assertFalse(im1.info.get("progressive"))
        self.assertFalse(im2.info.get("progressive"))
        self.assertTrue(im3.info.get("progressive"))

        self.assert_image_equal(im1, im3)
        self.assertGreaterEqual(im1.bytes, im3.bytes)

    def test_progressive_large_buffer(self):
        f = self.tempfile("temp.jpg")
        # this requires ~ 1.5x Image.MAXBLOCK
        im = Image.new("RGB", (4096, 4096), 0xFF3333)
        im.save(f, format="JPEG", progressive=True)

    def test_progressive_large_buffer_highest_quality(self):
        f = self.tempfile("temp.jpg")
        im = self.gen_random_image((255, 255))
        # this requires more bytes than pixels in the image
        im.save(f, format="JPEG", progressive=True, quality=100)

    def test_progressive_cmyk_buffer(self):
        # Issue 2272, quality 90 cmyk image is tripping the large buffer bug.
        f = BytesIO()
        im = self.gen_random_image((256, 256), "CMYK")
        im.save(f, format="JPEG", progressive=True, quality=94)

    def test_large_exif(self):
        # https://github.com/python-pillow/Pillow/issues/148
        f = self.tempfile("temp.jpg")
        im = hopper()
        im.save(f, "JPEG", quality=90, exif=b"1" * 65532)

    def test_exif_typeerror(self):
        im = Image.open("Tests/images/exif_typeerror.jpg")
        # Should not raise a TypeError
        im._getexif()

    def test_exif_gps(self):
        # Arrange
        im = Image.open("Tests/images/exif_gps.jpg")
        gps_index = 34853
        expected_exif_gps = {
            0: b"\x00\x00\x00\x01",
            2: (4294967295, 1),
            5: b"\x01",
            30: 65535,
            29: "1999:99:99 99:99:99",
        }

        # Act
        exif = im._getexif()

        # Assert
        self.assertEqual(exif[gps_index], expected_exif_gps)

    def test_exif_rollback(self):
        # rolling back exif support in 3.1 to pre-3.0 formatting.
        # expected from 2.9, with b/u qualifiers switched for 3.2 compatibility
        # this test passes on 2.9 and 3.1, but not 3.0
        expected_exif = {
            34867: 4294967295,
            258: (24, 24, 24),
            36867: "2099:09:29 10:10:10",
            34853: {
                0: b"\x00\x00\x00\x01",
                2: (4294967295, 1),
                5: b"\x01",
                30: 65535,
                29: "1999:99:99 99:99:99",
            },
            296: 65535,
            34665: 185,
            41994: 65535,
            514: 4294967295,
            271: "Make",
            272: "XXX-XXX",
            305: "PIL",
            42034: ((1, 1), (1, 1), (1, 1), (1, 1)),
            42035: "LensMake",
            34856: b"\xaa\xaa\xaa\xaa\xaa\xaa",
            282: (4294967295, 1),
            33434: (4294967295, 1),
        }

        im = Image.open("Tests/images/exif_gps.jpg")
        exif = im._getexif()

        for tag, value in expected_exif.items():
            self.assertEqual(value, exif[tag])

    def test_exif_gps_typeerror(self):
        im = Image.open("Tests/images/exif_gps_typeerror.jpg")

        # Should not raise a TypeError
        im._getexif()

    def test_progressive_compat(self):
        im1 = self.roundtrip(hopper())
        self.assertFalse(im1.info.get("progressive"))
        self.assertFalse(im1.info.get("progression"))

        im2 = self.roundtrip(hopper(), progressive=0)
        im3 = self.roundtrip(hopper(), progression=0)  # compatibility
        self.assertFalse(im2.info.get("progressive"))
        self.assertFalse(im2.info.get("progression"))
        self.assertFalse(im3.info.get("progressive"))
        self.assertFalse(im3.info.get("progression"))

        im2 = self.roundtrip(hopper(), progressive=1)
        im3 = self.roundtrip(hopper(), progression=1)  # compatibility
        self.assert_image_equal(im1, im2)
        self.assert_image_equal(im1, im3)
        self.assertTrue(im2.info.get("progressive"))
        self.assertTrue(im2.info.get("progression"))
        self.assertTrue(im3.info.get("progressive"))
        self.assertTrue(im3.info.get("progression"))

    def test_quality(self):
        im1 = self.roundtrip(hopper())
        im2 = self.roundtrip(hopper(), quality=50)
        self.assert_image(im1, im2.mode, im2.size)
        self.assertGreaterEqual(im1.bytes, im2.bytes)

    def test_smooth(self):
        im1 = self.roundtrip(hopper())
        im2 = self.roundtrip(hopper(), smooth=100)
        self.assert_image(im1, im2.mode, im2.size)

    def test_subsampling(self):
        def getsampling(im):
            layer = im.layer
            return layer[0][1:3] + layer[1][1:3] + layer[2][1:3]

        # experimental API
        im = self.roundtrip(hopper(), subsampling=-1)  # default
        self.assertEqual(getsampling(im), (2, 2, 1, 1, 1, 1))
        im = self.roundtrip(hopper(), subsampling=0)  # 4:4:4
        self.assertEqual(getsampling(im), (1, 1, 1, 1, 1, 1))
        im = self.roundtrip(hopper(), subsampling=1)  # 4:2:2
        self.assertEqual(getsampling(im), (2, 1, 1, 1, 1, 1))
        im = self.roundtrip(hopper(), subsampling=2)  # 4:2:0
        self.assertEqual(getsampling(im), (2, 2, 1, 1, 1, 1))
        im = self.roundtrip(hopper(), subsampling=3)  # default (undefined)
        self.assertEqual(getsampling(im), (2, 2, 1, 1, 1, 1))

        im = self.roundtrip(hopper(), subsampling="4:4:4")
        self.assertEqual(getsampling(im), (1, 1, 1, 1, 1, 1))
        im = self.roundtrip(hopper(), subsampling="4:2:2")
        self.assertEqual(getsampling(im), (2, 1, 1, 1, 1, 1))
        im = self.roundtrip(hopper(), subsampling="4:2:0")
        self.assertEqual(getsampling(im), (2, 2, 1, 1, 1, 1))
        im = self.roundtrip(hopper(), subsampling="4:1:1")
        self.assertEqual(getsampling(im), (2, 2, 1, 1, 1, 1))

        self.assertRaises(TypeError, self.roundtrip, hopper(), subsampling="1:1:1")

    def test_exif(self):
        im = Image.open("Tests/images/pil_sample_rgb.jpg")
        info = im._getexif()
        self.assertEqual(info[305], "Adobe Photoshop CS Macintosh")

    def test_mp(self):
        im = Image.open("Tests/images/pil_sample_rgb.jpg")
        self.assertIsNone(im._getmp())

    def test_quality_keep(self):
        # RGB
        im = Image.open("Tests/images/hopper.jpg")
        f = self.tempfile("temp.jpg")
        im.save(f, quality="keep")
        # Grayscale
        im = Image.open("Tests/images/hopper_gray.jpg")
        f = self.tempfile("temp.jpg")
        im.save(f, quality="keep")
        # CMYK
        im = Image.open("Tests/images/pil_sample_cmyk.jpg")
        f = self.tempfile("temp.jpg")
        im.save(f, quality="keep")

    def test_junk_jpeg_header(self):
        # https://github.com/python-pillow/Pillow/issues/630
        filename = "Tests/images/junk_jpeg_header.jpg"
        Image.open(filename)

    def test_ff00_jpeg_header(self):
        filename = "Tests/images/jpeg_ff00_header.jpg"
        Image.open(filename)

    def test_truncated_jpeg_should_read_all_the_data(self):
        filename = "Tests/images/truncated_jpeg.jpg"
        ImageFile.LOAD_TRUNCATED_IMAGES = True
        im = Image.open(filename)
        im.load()
        ImageFile.LOAD_TRUNCATED_IMAGES = False
        self.assertIsNotNone(im.getbbox())

    def test_truncated_jpeg_throws_IOError(self):
        filename = "Tests/images/truncated_jpeg.jpg"
        im = Image.open(filename)

        with self.assertRaises(IOError):
            im.load()

        # Test that the error is raised if loaded a second time
        with self.assertRaises(IOError):
            im.load()

    def _n_qtables_helper(self, n, test_file):
        im = Image.open(test_file)
        f = self.tempfile("temp.jpg")
        im.save(f, qtables=[[n] * 64] * n)
        im = Image.open(f)
        self.assertEqual(len(im.quantization), n)
        reloaded = self.roundtrip(im, qtables="keep")
        self.assertEqual(im.quantization, reloaded.quantization)

    def test_qtables(self):
        im = Image.open("Tests/images/hopper.jpg")
        qtables = im.quantization
        reloaded = self.roundtrip(im, qtables=qtables, subsampling=0)
        self.assertEqual(im.quantization, reloaded.quantization)
        self.assert_image_similar(im, self.roundtrip(im, qtables="web_low"), 30)
        self.assert_image_similar(im, self.roundtrip(im, qtables="web_high"), 30)
        self.assert_image_similar(im, self.roundtrip(im, qtables="keep"), 30)

        # valid bounds for baseline qtable
        bounds_qtable = [int(s) for s in ("255 1 " * 32).split(None)]
        self.roundtrip(im, qtables=[bounds_qtable])

        # values from wizard.txt in jpeg9-a src package.
        standard_l_qtable = [
            int(s)
            for s in """
            16  11  10  16  24  40  51  61
            12  12  14  19  26  58  60  55
            14  13  16  24  40  57  69  56
            14  17  22  29  51  87  80  62
            18  22  37  56  68 109 103  77
            24  35  55  64  81 104 113  92
            49  64  78  87 103 121 120 101
            72  92  95  98 112 100 103  99
            """.split(
                None
            )
        ]

        standard_chrominance_qtable = [
            int(s)
            for s in """
            17  18  24  47  99  99  99  99
            18  21  26  66  99  99  99  99
            24  26  56  99  99  99  99  99
            47  66  99  99  99  99  99  99
            99  99  99  99  99  99  99  99
            99  99  99  99  99  99  99  99
            99  99  99  99  99  99  99  99
            99  99  99  99  99  99  99  99
            """.split(
                None
            )
        ]
        # list of qtable lists
        self.assert_image_similar(
            im,
            self.roundtrip(
                im, qtables=[standard_l_qtable, standard_chrominance_qtable]
            ),
            30,
        )

        # tuple of qtable lists
        self.assert_image_similar(
            im,
            self.roundtrip(
                im, qtables=(standard_l_qtable, standard_chrominance_qtable)
            ),
            30,
        )

        # dict of qtable lists
        self.assert_image_similar(
            im,
            self.roundtrip(
                im, qtables={0: standard_l_qtable, 1: standard_chrominance_qtable}
            ),
            30,
        )

        self._n_qtables_helper(1, "Tests/images/hopper_gray.jpg")
        self._n_qtables_helper(1, "Tests/images/pil_sample_rgb.jpg")
        self._n_qtables_helper(2, "Tests/images/pil_sample_rgb.jpg")
        self._n_qtables_helper(3, "Tests/images/pil_sample_rgb.jpg")
        self._n_qtables_helper(1, "Tests/images/pil_sample_cmyk.jpg")
        self._n_qtables_helper(2, "Tests/images/pil_sample_cmyk.jpg")
        self._n_qtables_helper(3, "Tests/images/pil_sample_cmyk.jpg")
        self._n_qtables_helper(4, "Tests/images/pil_sample_cmyk.jpg")

        # not a sequence
        self.assertRaises(ValueError, self.roundtrip, im, qtables="a")
        # sequence wrong length
        self.assertRaises(ValueError, self.roundtrip, im, qtables=[])
        # sequence wrong length
        self.assertRaises(ValueError, self.roundtrip, im, qtables=[1, 2, 3, 4, 5])

        # qtable entry not a sequence
        self.assertRaises(ValueError, self.roundtrip, im, qtables=[1])
        # qtable entry has wrong number of items
        self.assertRaises(ValueError, self.roundtrip, im, qtables=[[1, 2, 3, 4]])

    @unittest.skipUnless(djpeg_available(), "djpeg not available")
    def test_load_djpeg(self):
        img = Image.open(TEST_FILE)
        img.load_djpeg()
        self.assert_image_similar(img, Image.open(TEST_FILE), 0)

    @unittest.skipUnless(cjpeg_available(), "cjpeg not available")
    def test_save_cjpeg(self):
        img = Image.open(TEST_FILE)

        tempfile = self.tempfile("temp.jpg")
        JpegImagePlugin._save_cjpeg(img, 0, tempfile)
        # Default save quality is 75%, so a tiny bit of difference is alright
        self.assert_image_similar(img, Image.open(tempfile), 17)

    def test_no_duplicate_0x1001_tag(self):
        # Arrange
        from PIL import ExifTags

        tag_ids = {v: k for k, v in ExifTags.TAGS.items()}

        # Assert
        self.assertEqual(tag_ids["RelatedImageWidth"], 0x1001)
        self.assertEqual(tag_ids["RelatedImageLength"], 0x1002)

    def test_MAXBLOCK_scaling(self):
        im = self.gen_random_image((512, 512))
        f = self.tempfile("temp.jpeg")
        im.save(f, quality=100, optimize=True)

        reloaded = Image.open(f)

        # none of these should crash
        reloaded.save(f, quality="keep")
        reloaded.save(f, quality="keep", progressive=True)
        reloaded.save(f, quality="keep", optimize=True)

    def test_bad_mpo_header(self):
        """ Treat unknown MPO as JPEG """
        # Arrange

        # Act
        # Shouldn't raise error
        fn = "Tests/images/sugarshack_bad_mpo_header.jpg"
        im = self.assert_warning(UserWarning, Image.open, fn)

        # Assert
        self.assertEqual(im.format, "JPEG")

    def test_save_correct_modes(self):
        out = BytesIO()
        for mode in ["1", "L", "RGB", "RGBX", "CMYK", "YCbCr"]:
            img = Image.new(mode, (20, 20))
            img.save(out, "JPEG")

    def test_save_wrong_modes(self):
        # ref https://github.com/python-pillow/Pillow/issues/2005
        out = BytesIO()
        for mode in ["LA", "La", "RGBA", "RGBa", "P"]:
            img = Image.new(mode, (20, 20))
            self.assertRaises(IOError, img.save, out, "JPEG")

    def test_save_tiff_with_dpi(self):
        # Arrange
        outfile = self.tempfile("temp.tif")
        im = Image.open("Tests/images/hopper.tif")

        # Act
        im.save(outfile, "JPEG", dpi=im.info["dpi"])

        # Assert
        reloaded = Image.open(outfile)
        reloaded.load()
        self.assertEqual(im.info["dpi"], reloaded.info["dpi"])

    def test_load_dpi_rounding(self):
        # Round up
        im = Image.open("Tests/images/iptc_roundUp.jpg")
        self.assertEqual(im.info["dpi"], (44, 44))

        # Round down
        im = Image.open("Tests/images/iptc_roundDown.jpg")
        self.assertEqual(im.info["dpi"], (2, 2))

    def test_save_dpi_rounding(self):
        outfile = self.tempfile("temp.jpg")
        im = Image.open("Tests/images/hopper.jpg")

        im.save(outfile, dpi=(72.2, 72.2))
        reloaded = Image.open(outfile)
        self.assertEqual(reloaded.info["dpi"], (72, 72))

        im.save(outfile, dpi=(72.8, 72.8))
        reloaded = Image.open(outfile)
        self.assertEqual(reloaded.info["dpi"], (73, 73))

    def test_dpi_tuple_from_exif(self):
        # Arrange
        # This Photoshop CC 2017 image has DPI in EXIF not metadata
        # EXIF XResolution is (2000000, 10000)
        im = Image.open("Tests/images/photoshop-200dpi.jpg")

        # Act / Assert
        self.assertEqual(im.info.get("dpi"), (200, 200))

    def test_dpi_int_from_exif(self):
        # Arrange
        # This image has DPI in EXIF not metadata
        # EXIF XResolution is 72
        im = Image.open("Tests/images/exif-72dpi-int.jpg")

        # Act / Assert
        self.assertEqual(im.info.get("dpi"), (72, 72))

    def test_dpi_from_dpcm_exif(self):
        # Arrange
        # This is photoshop-200dpi.jpg with EXIF resolution unit set to cm:
        # exiftool -exif:ResolutionUnit=cm photoshop-200dpi.jpg
        im = Image.open("Tests/images/exif-200dpcm.jpg")

        # Act / Assert
        self.assertEqual(im.info.get("dpi"), (508, 508))

    def test_dpi_exif_zero_division(self):
        # Arrange
        # This is photoshop-200dpi.jpg with EXIF resolution set to 0/0:
        # exiftool -XResolution=0/0 -YResolution=0/0 photoshop-200dpi.jpg
        im = Image.open("Tests/images/exif-dpi-zerodivision.jpg")

        # Act / Assert
        # This should return the default, and not raise a ZeroDivisionError
        self.assertEqual(im.info.get("dpi"), (72, 72))

    def test_no_dpi_in_exif(self):
        # Arrange
        # This is photoshop-200dpi.jpg with resolution removed from EXIF:
        # exiftool "-*resolution*"= photoshop-200dpi.jpg
        im = Image.open("Tests/images/no-dpi-in-exif.jpg")

        # Act / Assert
        # "When the image resolution is unknown, 72 [dpi] is designated."
        # http://www.exiv2.org/tags.html
        self.assertEqual(im.info.get("dpi"), (72, 72))

    def test_invalid_exif(self):
        # This is no-dpi-in-exif with the tiff header of the exif block
        # hexedited from MM * to FF FF FF FF
        im = Image.open("Tests/images/invalid-exif.jpg")

        # This should return the default, and not a SyntaxError or
        # OSError for unidentified image.
        self.assertEqual(im.info.get("dpi"), (72, 72))

    def test_ifd_offset_exif(self):
        # Arrange
        # This image has been manually hexedited to have an IFD offset of 10,
        # in contrast to normal 8
        im = Image.open("Tests/images/exif-ifd-offset.jpg")

        # Act / Assert
        self.assertEqual(im._getexif()[306], "2017:03:13 23:03:09")

    def test_photoshop(self):
        im = Image.open("Tests/images/photoshop-200dpi.jpg")
        self.assertEqual(
            im.info["photoshop"][0x03ED],
            {
                "XResolution": 200.0,
                "DisplayedUnitsX": 1,
                "YResolution": 200.0,
                "DisplayedUnitsY": 1,
            },
        )

        # This image does not contain a Photoshop header string
        im = Image.open("Tests/images/app13.jpg")
        self.assertNotIn("photoshop", im.info)


@unittest.skipUnless(sys.platform.startswith("win32"), "Windows only")
class TestFileCloseW32(PillowTestCase):
    def setUp(self):
        if "jpeg_encoder" not in codecs or "jpeg_decoder" not in codecs:
            self.skipTest("jpeg support not available")

    def test_fd_leak(self):
        tmpfile = self.tempfile("temp.jpg")

        with Image.open("Tests/images/hopper.jpg") as im:
            im.save(tmpfile)

        im = Image.open(tmpfile)
        fp = im.fp
        self.assertFalse(fp.closed)
        self.assertRaises(WindowsError, os.remove, tmpfile)
        im.load()
        self.assertTrue(fp.closed)
        # this should not fail, as load should have closed the file.
        os.remove(tmpfile)



from PIL import Image, PixarImagePlugin

from .helper import PillowTestCase, hopper

TEST_FILE = "Tests/images/hopper.pxr"


class TestFilePixar(PillowTestCase):
    def test_sanity(self):
        im = Image.open(TEST_FILE)
        im.load()
        self.assertEqual(im.mode, "RGB")
        self.assertEqual(im.size, (128, 128))
        self.assertEqual(im.format, "PIXAR")
        self.assertIsNone(im.get_format_mimetype())

        im2 = hopper()
        self.assert_image_similar(im, im2, 4.8)

    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, PixarImagePlugin.PixarImageFile, invalid_file)



from __future__ import print_function

import distutils.version
import io
import itertools
import logging
import os
from collections import namedtuple
from ctypes import c_float

from PIL import Image, TiffImagePlugin, TiffTags, features
from PIL._util import py3

from .helper import PillowTestCase, hopper

logger = logging.getLogger(__name__)


class LibTiffTestCase(PillowTestCase):
    def setUp(self):
        if not features.check("libtiff"):
            self.skipTest("tiff support not available")

    def _assert_noerr(self, im):
        """Helper tests that assert basic sanity about the g4 tiff reading"""
        # 1 bit
        self.assertEqual(im.mode, "1")

        # Does the data actually load
        im.load()
        im.getdata()

        try:
            self.assertEqual(im._compression, "group4")
        except AttributeError:
            print("No _compression")
            print(dir(im))

        # can we write it back out, in a different form.
        out = self.tempfile("temp.png")
        im.save(out)

        out_bytes = io.BytesIO()
        im.save(out_bytes, format="tiff", compression="group4")


class TestFileLibTiff(LibTiffTestCase):
    def test_g4_tiff(self):
        """Test the ordinary file path load path"""

        test_file = "Tests/images/hopper_g4_500.tif"
        im = Image.open(test_file)

        self.assertEqual(im.size, (500, 500))
        self._assert_noerr(im)

    def test_g4_large(self):
        test_file = "Tests/images/pport_g4.tif"
        im = Image.open(test_file)
        self._assert_noerr(im)

    def test_g4_tiff_file(self):
        """Testing the string load path"""

        test_file = "Tests/images/hopper_g4_500.tif"
        with open(test_file, "rb") as f:
            im = Image.open(f)

            self.assertEqual(im.size, (500, 500))
            self._assert_noerr(im)

    def test_g4_tiff_bytesio(self):
        """Testing the stringio loading code path"""
        test_file = "Tests/images/hopper_g4_500.tif"
        s = io.BytesIO()
        with open(test_file, "rb") as f:
            s.write(f.read())
            s.seek(0)
        im = Image.open(s)

        self.assertEqual(im.size, (500, 500))
        self._assert_noerr(im)

    def test_g4_non_disk_file_object(self):
        """Testing loading from non-disk non-BytesIO file object"""
        test_file = "Tests/images/hopper_g4_500.tif"
        s = io.BytesIO()
        with open(test_file, "rb") as f:
            s.write(f.read())
            s.seek(0)
        r = io.BufferedReader(s)
        im = Image.open(r)

        self.assertEqual(im.size, (500, 500))
        self._assert_noerr(im)

    def test_g4_eq_png(self):
        """ Checking that we're actually getting the data that we expect"""
        png = Image.open("Tests/images/hopper_bw_500.png")
        g4 = Image.open("Tests/images/hopper_g4_500.tif")

        self.assert_image_equal(g4, png)

    # see https://github.com/python-pillow/Pillow/issues/279
    def test_g4_fillorder_eq_png(self):
        """ Checking that we're actually getting the data that we expect"""
        png = Image.open("Tests/images/g4-fillorder-test.png")
        g4 = Image.open("Tests/images/g4-fillorder-test.tif")

        self.assert_image_equal(g4, png)

    def test_g4_write(self):
        """Checking to see that the saved image is the same as what we wrote"""
        test_file = "Tests/images/hopper_g4_500.tif"
        orig = Image.open(test_file)

        out = self.tempfile("temp.tif")
        rot = orig.transpose(Image.ROTATE_90)
        self.assertEqual(rot.size, (500, 500))
        rot.save(out)

        reread = Image.open(out)
        self.assertEqual(reread.size, (500, 500))
        self._assert_noerr(reread)
        self.assert_image_equal(reread, rot)
        self.assertEqual(reread.info["compression"], "group4")

        self.assertEqual(reread.info["compression"], orig.info["compression"])

        self.assertNotEqual(orig.tobytes(), reread.tobytes())

    def test_adobe_deflate_tiff(self):
        test_file = "Tests/images/tiff_adobe_deflate.tif"
        im = Image.open(test_file)

        self.assertEqual(im.mode, "RGB")
        self.assertEqual(im.size, (278, 374))
        self.assertEqual(im.tile[0][:3], ("libtiff", (0, 0, 278, 374), 0))
        im.load()

        self.assert_image_equal_tofile(im, "Tests/images/tiff_adobe_deflate.png")

    def test_write_metadata(self):
        """ Test metadata writing through libtiff """
        for legacy_api in [False, True]:
            img = Image.open("Tests/images/hopper_g4.tif")
            f = self.tempfile("temp.tiff")

            img.save(f, tiffinfo=img.tag)

            if legacy_api:
                original = img.tag.named()
            else:
                original = img.tag_v2.named()

            # PhotometricInterpretation is set from SAVE_INFO,
            # not the original image.
            ignored = [
                "StripByteCounts",
                "RowsPerStrip",
                "PageNumber",
                "PhotometricInterpretation",
            ]

            loaded = Image.open(f)
            if legacy_api:
                reloaded = loaded.tag.named()
            else:
                reloaded = loaded.tag_v2.named()

            for tag, value in itertools.chain(reloaded.items(), original.items()):
                if tag not in ignored:
                    val = original[tag]
                    if tag.endswith("Resolution"):
                        if legacy_api:
                            self.assertEqual(
                                c_float(val[0][0] / val[0][1]).value,
                                c_float(value[0][0] / value[0][1]).value,
                                msg="%s didn't roundtrip" % tag,
                            )
                        else:
                            self.assertEqual(
                                c_float(val).value,
                                c_float(value).value,
                                msg="%s didn't roundtrip" % tag,
                            )
                    else:
                        self.assertEqual(val, value, msg="%s didn't roundtrip" % tag)

            # https://github.com/python-pillow/Pillow/issues/1561
            requested_fields = ["StripByteCounts", "RowsPerStrip", "StripOffsets"]
            for field in requested_fields:
                self.assertIn(field, reloaded, "%s not in metadata" % field)

    def test_additional_metadata(self):
        # these should not crash. Seriously dummy data, most of it doesn't make
        # any sense, so we're running up against limits where we're asking
        # libtiff to do stupid things.

        # Get the list of the ones that we should be able to write

        core_items = {
            tag: info
            for tag, info in ((s, TiffTags.lookup(s)) for s in TiffTags.LIBTIFF_CORE)
            if info.type is not None
        }

        # Exclude ones that have special meaning
        # that we're already testing them
        im = Image.open("Tests/images/hopper_g4.tif")
        for tag in im.tag_v2:
            try:
                del core_items[tag]
            except KeyError:
                pass

        # Type codes:
        #     2: "ascii",
        #     3: "short",
        #     4: "long",
        #     5: "rational",
        #     12: "double",
        # Type: dummy value
        values = {
            2: "test",
            3: 1,
            4: 2 ** 20,
            5: TiffImagePlugin.IFDRational(100, 1),
            12: 1.05,
        }

        new_ifd = TiffImagePlugin.ImageFileDirectory_v2()
        for tag, info in core_items.items():
            if info.length == 1:
                new_ifd[tag] = values[info.type]
            if info.length == 0:
                new_ifd[tag] = tuple(values[info.type] for _ in range(3))
            else:
                new_ifd[tag] = tuple(values[info.type] for _ in range(info.length))

        # Extra samples really doesn't make sense in this application.
        del new_ifd[338]

        out = self.tempfile("temp.tif")
        TiffImagePlugin.WRITE_LIBTIFF = True

        im.save(out, tiffinfo=new_ifd)

        TiffImagePlugin.WRITE_LIBTIFF = False

    def test_custom_metadata(self):
        tc = namedtuple("test_case", "value,type,supported_by_default")
        custom = {
            37000 + k: v
            for k, v in enumerate(
                [
                    tc(4, TiffTags.SHORT, True),
                    tc(123456789, TiffTags.LONG, True),
                    tc(-4, TiffTags.SIGNED_BYTE, False),
                    tc(-4, TiffTags.SIGNED_SHORT, False),
                    tc(-123456789, TiffTags.SIGNED_LONG, False),
                    tc(TiffImagePlugin.IFDRational(4, 7), TiffTags.RATIONAL, True),
                    tc(4.25, TiffTags.FLOAT, True),
                    tc(4.25, TiffTags.DOUBLE, True),
                    tc("custom tag value", TiffTags.ASCII, True),
                    tc(u"custom tag value", TiffTags.ASCII, True),
                    tc(b"custom tag value", TiffTags.BYTE, True),
                    tc((4, 5, 6), TiffTags.SHORT, True),
                    tc((123456789, 9, 34, 234, 219387, 92432323), TiffTags.LONG, True),
                    tc((-4, 9, 10), TiffTags.SIGNED_BYTE, False),
                    tc((-4, 5, 6), TiffTags.SIGNED_SHORT, False),
                    tc(
                        (-123456789, 9, 34, 234, 219387, -92432323),
                        TiffTags.SIGNED_LONG,
                        False,
                    ),
                    tc((4.25, 5.25), TiffTags.FLOAT, True),
                    tc((4.25, 5.25), TiffTags.DOUBLE, True),
                    # array of TIFF_BYTE requires bytes instead of tuple for backwards
                    # compatibility
                    tc(bytes([4]), TiffTags.BYTE, True),
                    tc(bytes((4, 9, 10)), TiffTags.BYTE, True),
                ]
            )
        }

        libtiff_version = TiffImagePlugin._libtiff_version()

        libtiffs = [False]
        if distutils.version.StrictVersion(
            libtiff_version
        ) >= distutils.version.StrictVersion("4.0"):
            libtiffs.append(True)

        for libtiff in libtiffs:
            TiffImagePlugin.WRITE_LIBTIFF = libtiff

            def check_tags(tiffinfo):
                im = hopper()

                out = self.tempfile("temp.tif")
                im.save(out, tiffinfo=tiffinfo)

                reloaded = Image.open(out)
                for tag, value in tiffinfo.items():
                    reloaded_value = reloaded.tag_v2[tag]
                    if (
                        isinstance(reloaded_value, TiffImagePlugin.IFDRational)
                        and libtiff
                    ):
                        # libtiff does not support real RATIONALS
                        self.assertAlmostEqual(float(reloaded_value), float(value))
                        continue

                    if libtiff and isinstance(value, bytes):
                        value = value.decode()

                    self.assertEqual(reloaded_value, value)

            # Test with types
            ifd = TiffImagePlugin.ImageFileDirectory_v2()
            for tag, tagdata in custom.items():
                ifd[tag] = tagdata.value
                ifd.tagtype[tag] = tagdata.type
            check_tags(ifd)

            # Test without types. This only works for some types, int for example are
            # always encoded as LONG and not SIGNED_LONG.
            check_tags(
                {
                    tag: tagdata.value
                    for tag, tagdata in custom.items()
                    if tagdata.supported_by_default
                }
            )
        TiffImagePlugin.WRITE_LIBTIFF = False

    def test_int_dpi(self):
        # issue #1765
        im = hopper("RGB")
        out = self.tempfile("temp.tif")
        TiffImagePlugin.WRITE_LIBTIFF = True
        im.save(out, dpi=(72, 72))
        TiffImagePlugin.WRITE_LIBTIFF = False
        reloaded = Image.open(out)
        self.assertEqual(reloaded.info["dpi"], (72.0, 72.0))

    def test_g3_compression(self):
        i = Image.open("Tests/images/hopper_g4_500.tif")
        out = self.tempfile("temp.tif")
        i.save(out, compression="group3")

        reread = Image.open(out)
        self.assertEqual(reread.info["compression"], "group3")
        self.assert_image_equal(reread, i)

    def test_little_endian(self):
        im = Image.open("Tests/images/16bit.deflate.tif")
        self.assertEqual(im.getpixel((0, 0)), 480)
        self.assertEqual(im.mode, "I;16")

        b = im.tobytes()
        # Bytes are in image native order (little endian)
        if py3:
            self.assertEqual(b[0], ord(b"\xe0"))
            self.assertEqual(b[1], ord(b"\x01"))
        else:
            self.assertEqual(b[0], b"\xe0")
            self.assertEqual(b[1], b"\x01")

        out = self.tempfile("temp.tif")
        # out = "temp.le.tif"
        im.save(out)
        reread = Image.open(out)

        self.assertEqual(reread.info["compression"], im.info["compression"])
        self.assertEqual(reread.getpixel((0, 0)), 480)
        # UNDONE - libtiff defaults to writing in native endian, so
        # on big endian, we'll get back mode = 'I;16B' here.

    def test_big_endian(self):
        im = Image.open("Tests/images/16bit.MM.deflate.tif")

        self.assertEqual(im.getpixel((0, 0)), 480)
        self.assertEqual(im.mode, "I;16B")

        b = im.tobytes()

        # Bytes are in image native order (big endian)
        if py3:
            self.assertEqual(b[0], ord(b"\x01"))
            self.assertEqual(b[1], ord(b"\xe0"))
        else:
            self.assertEqual(b[0], b"\x01")
            self.assertEqual(b[1], b"\xe0")

        out = self.tempfile("temp.tif")
        im.save(out)
        reread = Image.open(out)

        self.assertEqual(reread.info["compression"], im.info["compression"])
        self.assertEqual(reread.getpixel((0, 0)), 480)

    def test_g4_string_info(self):
        """Tests String data in info directory"""
        test_file = "Tests/images/hopper_g4_500.tif"
        orig = Image.open(test_file)

        out = self.tempfile("temp.tif")

        orig.tag[269] = "temp.tif"
        orig.save(out)

        reread = Image.open(out)
        self.assertEqual("temp.tif", reread.tag_v2[269])
        self.assertEqual("temp.tif", reread.tag[269][0])

    def test_12bit_rawmode(self):
        """ Are we generating the same interpretation
        of the image as Imagemagick is? """
        TiffImagePlugin.READ_LIBTIFF = True
        im = Image.open("Tests/images/12bit.cropped.tif")
        im.load()
        TiffImagePlugin.READ_LIBTIFF = False
        # to make the target --
        # convert 12bit.cropped.tif -depth 16 tmp.tif
        # convert tmp.tif -evaluate RightShift 4 12in16bit2.tif
        # imagemagick will auto scale so that a 12bit FFF is 16bit FFF0,
        # so we need to unshift so that the integer values are the same.

        self.assert_image_equal_tofile(im, "Tests/images/12in16bit.tif")

    def test_blur(self):
        # test case from irc, how to do blur on b/w image
        # and save to compressed tif.
        from PIL import ImageFilter

        out = self.tempfile("temp.tif")
        im = Image.open("Tests/images/pport_g4.tif")
        im = im.convert("L")

        im = im.filter(ImageFilter.GaussianBlur(4))
        im.save(out, compression="tiff_adobe_deflate")

        im2 = Image.open(out)
        im2.load()

        self.assert_image_equal(im, im2)

    def test_compressions(self):
        # Test various tiff compressions and assert similar image content but reduced
        # file sizes.
        im = hopper("RGB")
        out = self.tempfile("temp.tif")
        im.save(out)
        size_raw = os.path.getsize(out)

        for compression in ("packbits", "tiff_lzw"):
            im.save(out, compression=compression)
            size_compressed = os.path.getsize(out)
            im2 = Image.open(out)
            self.assert_image_equal(im, im2)

        im.save(out, compression="jpeg")
        size_jpeg = os.path.getsize(out)
        im2 = Image.open(out)
        self.assert_image_similar(im, im2, 30)

        im.save(out, compression="jpeg", quality=30)
        size_jpeg_30 = os.path.getsize(out)
        im3 = Image.open(out)
        self.assert_image_similar(im2, im3, 30)

        self.assertGreater(size_raw, size_compressed)
        self.assertGreater(size_compressed, size_jpeg)
        self.assertGreater(size_jpeg, size_jpeg_30)

    def test_quality(self):
        im = hopper("RGB")
        out = self.tempfile("temp.tif")

        self.assertRaises(ValueError, im.save, out, compression="tiff_lzw", quality=50)
        self.assertRaises(ValueError, im.save, out, compression="jpeg", quality=-1)
        self.assertRaises(ValueError, im.save, out, compression="jpeg", quality=101)
        self.assertRaises(ValueError, im.save, out, compression="jpeg", quality="good")
        im.save(out, compression="jpeg", quality=0)
        im.save(out, compression="jpeg", quality=100)

    def test_cmyk_save(self):
        im = hopper("CMYK")
        out = self.tempfile("temp.tif")

        im.save(out, compression="tiff_adobe_deflate")
        im2 = Image.open(out)
        self.assert_image_equal(im, im2)

    def xtest_bw_compression_w_rgb(self):
        """ This test passes, but when running all tests causes a failure due
            to output on stderr from the error thrown by libtiff. We need to
            capture that but not now"""

        im = hopper("RGB")
        out = self.tempfile("temp.tif")

        self.assertRaises(IOError, im.save, out, compression="tiff_ccitt")
        self.assertRaises(IOError, im.save, out, compression="group3")
        self.assertRaises(IOError, im.save, out, compression="group4")

    def test_fp_leak(self):
        im = Image.open("Tests/images/hopper_g4_500.tif")
        fn = im.fp.fileno()

        os.fstat(fn)
        im.load()  # this should close it.
        self.assertRaises(OSError, os.fstat, fn)
        im = None  # this should force even more closed.
        self.assertRaises(OSError, os.fstat, fn)
        self.assertRaises(OSError, os.close, fn)

    def test_multipage(self):
        # issue #862
        TiffImagePlugin.READ_LIBTIFF = True
        im = Image.open("Tests/images/multipage.tiff")
        # file is a multipage tiff,  10x10 green, 10x10 red, 20x20 blue

        im.seek(0)
        self.assertEqual(im.size, (10, 10))
        self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 128, 0))
        self.assertTrue(im.tag.next)

        im.seek(1)
        self.assertEqual(im.size, (10, 10))
        self.assertEqual(im.convert("RGB").getpixel((0, 0)), (255, 0, 0))
        self.assertTrue(im.tag.next)

        im.seek(2)
        self.assertFalse(im.tag.next)
        self.assertEqual(im.size, (20, 20))
        self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 0, 255))

        TiffImagePlugin.READ_LIBTIFF = False

    def test_multipage_nframes(self):
        # issue #862
        TiffImagePlugin.READ_LIBTIFF = True
        im = Image.open("Tests/images/multipage.tiff")
        frames = im.n_frames
        self.assertEqual(frames, 3)
        for _ in range(frames):
            im.seek(0)
            # Should not raise ValueError: I/O operation on closed file
            im.load()

        TiffImagePlugin.READ_LIBTIFF = False

    def test__next(self):
        TiffImagePlugin.READ_LIBTIFF = True
        im = Image.open("Tests/images/hopper.tif")
        self.assertFalse(im.tag.next)
        im.load()
        self.assertFalse(im.tag.next)

    def test_4bit(self):
        # Arrange
        test_file = "Tests/images/hopper_gray_4bpp.tif"
        original = hopper("L")

        # Act
        TiffImagePlugin.READ_LIBTIFF = True
        im = Image.open(test_file)
        TiffImagePlugin.READ_LIBTIFF = False

        # Assert
        self.assertEqual(im.size, (128, 128))
        self.assertEqual(im.mode, "L")
        self.assert_image_similar(im, original, 7.3)

    def test_gray_semibyte_per_pixel(self):
        test_files = (
            (
                24.8,  # epsilon
                (  # group
                    "Tests/images/tiff_gray_2_4_bpp/hopper2.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2I.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2R.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2IR.tif",
                ),
            ),
            (
                7.3,  # epsilon
                (  # group
                    "Tests/images/tiff_gray_2_4_bpp/hopper4.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4I.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4R.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4IR.tif",
                ),
            ),
        )
        original = hopper("L")
        for epsilon, group in test_files:
            im = Image.open(group[0])
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.mode, "L")
            self.assert_image_similar(im, original, epsilon)
            for file in group[1:]:
                im2 = Image.open(file)
                self.assertEqual(im2.size, (128, 128))
                self.assertEqual(im2.mode, "L")
                self.assert_image_equal(im, im2)

    def test_save_bytesio(self):
        # PR 1011
        # Test TIFF saving to io.BytesIO() object.

        TiffImagePlugin.WRITE_LIBTIFF = True
        TiffImagePlugin.READ_LIBTIFF = True

        # Generate test image
        pilim = hopper()

        def save_bytesio(compression=None):

            buffer_io = io.BytesIO()
            pilim.save(buffer_io, format="tiff", compression=compression)
            buffer_io.seek(0)

            pilim_load = Image.open(buffer_io)
            self.assert_image_similar(pilim, pilim_load, 0)

        save_bytesio()
        save_bytesio("raw")
        save_bytesio("packbits")
        save_bytesio("tiff_lzw")

        TiffImagePlugin.WRITE_LIBTIFF = False
        TiffImagePlugin.READ_LIBTIFF = False

    def test_crashing_metadata(self):
        # issue 1597
        im = Image.open("Tests/images/rdf.tif")
        out = self.tempfile("temp.tif")

        TiffImagePlugin.WRITE_LIBTIFF = True
        # this shouldn't crash
        im.save(out, format="TIFF")
        TiffImagePlugin.WRITE_LIBTIFF = False

    def test_page_number_x_0(self):
        # Issue 973
        # Test TIFF with tag 297 (Page Number) having value of 0 0.
        # The first number is the current page number.
        # The second is the total number of pages, zero means not available.
        outfile = self.tempfile("temp.tif")
        # Created by printing a page in Chrome to PDF, then:
        # /usr/bin/gs -q -sDEVICE=tiffg3 -sOutputFile=total-pages-zero.tif
        # -dNOPAUSE /tmp/test.pdf -c quit
        infile = "Tests/images/total-pages-zero.tif"
        im = Image.open(infile)
        # Should not divide by zero
        im.save(outfile)

    def test_fd_duplication(self):
        # https://github.com/python-pillow/Pillow/issues/1651

        tmpfile = self.tempfile("temp.tif")
        with open(tmpfile, "wb") as f:
            with open("Tests/images/g4-multi.tiff", "rb") as src:
                f.write(src.read())

        im = Image.open(tmpfile)
        im.n_frames
        im.close()
        # Should not raise PermissionError.
        os.remove(tmpfile)

    def test_read_icc(self):
        with Image.open("Tests/images/hopper.iccprofile.tif") as img:
            icc = img.info.get("icc_profile")
            self.assertIsNotNone(icc)
        TiffImagePlugin.READ_LIBTIFF = True
        with Image.open("Tests/images/hopper.iccprofile.tif") as img:
            icc_libtiff = img.info.get("icc_profile")
            self.assertIsNotNone(icc_libtiff)
        TiffImagePlugin.READ_LIBTIFF = False
        self.assertEqual(icc, icc_libtiff)

    def test_multipage_compression(self):
        im = Image.open("Tests/images/compression.tif")

        im.seek(0)
        self.assertEqual(im._compression, "tiff_ccitt")
        self.assertEqual(im.size, (10, 10))

        im.seek(1)
        self.assertEqual(im._compression, "packbits")
        self.assertEqual(im.size, (10, 10))
        im.load()

        im.seek(0)
        self.assertEqual(im._compression, "tiff_ccitt")
        self.assertEqual(im.size, (10, 10))
        im.load()

    def test_save_tiff_with_jpegtables(self):
        # Arrange
        outfile = self.tempfile("temp.tif")

        # Created with ImageMagick: convert hopper.jpg hopper_jpg.tif
        # Contains JPEGTables (347) tag
        infile = "Tests/images/hopper_jpg.tif"
        im = Image.open(infile)

        # Act / Assert
        # Should not raise UnicodeDecodeError or anything else
        im.save(outfile)

    def test_16bit_RGB_tiff(self):
        im = Image.open("Tests/images/tiff_16bit_RGB.tiff")

        self.assertEqual(im.mode, "RGB")
        self.assertEqual(im.size, (100, 40))
        self.assertEqual(
            im.tile,
            [
                (
                    "libtiff",
                    (0, 0, 100, 40),
                    0,
                    ("RGB;16N", "tiff_adobe_deflate", False, 8),
                )
            ],
        )
        im.load()

        self.assert_image_equal_tofile(im, "Tests/images/tiff_16bit_RGB_target.png")

    def test_16bit_RGBa_tiff(self):
        im = Image.open("Tests/images/tiff_16bit_RGBa.tiff")

        self.assertEqual(im.mode, "RGBA")
        self.assertEqual(im.size, (100, 40))
        self.assertEqual(
            im.tile,
            [("libtiff", (0, 0, 100, 40), 0, ("RGBa;16N", "tiff_lzw", False, 38236))],
        )
        im.load()

        self.assert_image_equal_tofile(im, "Tests/images/tiff_16bit_RGBa_target.png")

    def test_gimp_tiff(self):
        # Read TIFF JPEG images from GIMP [@PIL168]

        codecs = dir(Image.core)
        if "jpeg_decoder" not in codecs:
            self.skipTest("jpeg support not available")

        filename = "Tests/images/pil168.tif"
        im = Image.open(filename)

        self.assertEqual(im.mode, "RGB")
        self.assertEqual(im.size, (256, 256))
        self.assertEqual(
            im.tile, [("libtiff", (0, 0, 256, 256), 0, ("RGB", "jpeg", False, 5122))]
        )
        im.load()

        self.assert_image_equal_tofile(im, "Tests/images/pil168.png")

    def test_sampleformat(self):
        # https://github.com/python-pillow/Pillow/issues/1466
        im = Image.open("Tests/images/copyleft.tiff")
        self.assertEqual(im.mode, "RGB")

        self.assert_image_equal_tofile(im, "Tests/images/copyleft.png", mode="RGB")

    def test_lzw(self):
        im = Image.open("Tests/images/hopper_lzw.tif")

        self.assertEqual(im.mode, "RGB")
        self.assertEqual(im.size, (128, 128))
        self.assertEqual(im.format, "TIFF")
        im2 = hopper()
        self.assert_image_similar(im, im2, 5)

    def test_strip_cmyk_jpeg(self):
        infile = "Tests/images/tiff_strip_cmyk_jpeg.tif"
        im = Image.open(infile)

        self.assert_image_similar_tofile(im, "Tests/images/pil_sample_cmyk.jpg", 0.5)

    def test_strip_cmyk_16l_jpeg(self):
        infile = "Tests/images/tiff_strip_cmyk_16l_jpeg.tif"
        im = Image.open(infile)

        self.assert_image_similar_tofile(im, "Tests/images/pil_sample_cmyk.jpg", 0.5)

    def test_strip_ycbcr_jpeg_2x2_sampling(self):
        infile = "Tests/images/tiff_strip_ycbcr_jpeg_2x2_sampling.tif"
        im = Image.open(infile)

        self.assert_image_similar_tofile(im, "Tests/images/flower.jpg", 0.5)

    def test_strip_ycbcr_jpeg_1x1_sampling(self):
        infile = "Tests/images/tiff_strip_ycbcr_jpeg_1x1_sampling.tif"
        im = Image.open(infile)

        self.assert_image_equal_tofile(im, "Tests/images/flower2.jpg")

    def test_tiled_cmyk_jpeg(self):
        infile = "Tests/images/tiff_tiled_cmyk_jpeg.tif"
        im = Image.open(infile)

        self.assert_image_similar_tofile(im, "Tests/images/pil_sample_cmyk.jpg", 0.5)

    def test_tiled_ycbcr_jpeg_1x1_sampling(self):
        infile = "Tests/images/tiff_tiled_ycbcr_jpeg_1x1_sampling.tif"
        im = Image.open(infile)

        self.assert_image_equal_tofile(im, "Tests/images/flower2.jpg")

    def test_tiled_ycbcr_jpeg_2x2_sampling(self):
        infile = "Tests/images/tiff_tiled_ycbcr_jpeg_2x2_sampling.tif"
        im = Image.open(infile)

        self.assert_image_similar_tofile(im, "Tests/images/flower.jpg", 0.5)

    def test_old_style_jpeg(self):
        infile = "Tests/images/old-style-jpeg-compression.tif"
        im = Image.open(infile)

        self.assert_image_equal_tofile(
            im, "Tests/images/old-style-jpeg-compression.png"
        )

    def test_no_rows_per_strip(self):
        # This image does not have a RowsPerStrip TIFF tag
        infile = "Tests/images/no_rows_per_strip.tif"
        im = Image.open(infile)
        im.load()
        self.assertEqual(im.size, (950, 975))

    def test_orientation(self):
        base_im = Image.open("Tests/images/g4_orientation_1.tif")

        for i in range(2, 9):
            im = Image.open("Tests/images/g4_orientation_" + str(i) + ".tif")
            im.load()

            self.assert_image_similar(base_im, im, 0.7)

    def test_sampleformat_not_corrupted(self):
        # Assert that a TIFF image with SampleFormat=UINT tag is not corrupted
        # when saving to a new file.
        # Pillow 6.0 fails with "OSError: cannot identify image file".
        import base64

        tiff = io.BytesIO(
            base64.b64decode(
                b"SUkqAAgAAAAPAP4ABAABAAAAAAAAAAABBAABAAAAAQAAAAEBBAABAAAAAQAA"
                b"AAIBAwADAAAAwgAAAAMBAwABAAAACAAAAAYBAwABAAAAAgAAABEBBAABAAAA"
                b"4AAAABUBAwABAAAAAwAAABYBBAABAAAAAQAAABcBBAABAAAACwAAABoBBQAB"
                b"AAAAyAAAABsBBQABAAAA0AAAABwBAwABAAAAAQAAACgBAwABAAAAAQAAAFMB"
                b"AwADAAAA2AAAAAAAAAAIAAgACAABAAAAAQAAAAEAAAABAAAAAQABAAEAAAB4"
                b"nGNgYAAAAAMAAQ=="
            )
        )
        out = io.BytesIO()
        with Image.open(tiff) as im:
            im.save(out, format="tiff")
        out.seek(0)
        with Image.open(out) as im:
            im.load()

    def test_realloc_overflow(self):
        TiffImagePlugin.READ_LIBTIFF = True
        with Image.open("Tests/images/tiff_overflow_rows_per_strip.tif") as im:
            with self.assertRaises(IOError) as e:
                im.load()

            # Assert that the error code is IMAGING_CODEC_MEMORY
            self.assertEqual(str(e.exception), "-9")
        TiffImagePlugin.READ_LIBTIFF = False



from PIL import GribStubImagePlugin, Image

from .helper import PillowTestCase, hopper

TEST_FILE = "Tests/images/WAlaska.wind.7days.grb"


class TestFileGribStub(PillowTestCase):
    def test_open(self):
        # Act
        im = Image.open(TEST_FILE)

        # Assert
        self.assertEqual(im.format, "GRIB")

        # Dummy data from the stub
        self.assertEqual(im.mode, "F")
        self.assertEqual(im.size, (1, 1))

    def test_invalid_file(self):
        # Arrange
        invalid_file = "Tests/images/flower.jpg"

        # Act / Assert
        self.assertRaises(
            SyntaxError, GribStubImagePlugin.GribStubImageFile, invalid_file
        )

    def test_load(self):
        # Arrange
        im = Image.open(TEST_FILE)

        # Act / Assert: stub cannot load without an implemented handler
        self.assertRaises(IOError, im.load)

    def test_save(self):
        # Arrange
        im = hopper()
        tmpfile = self.tempfile("temp.grib")

        # Act / Assert: stub cannot save without an implemented handler
        self.assertRaises(IOError, im.save, tmpfile)



from PIL import Image

from .helper import PillowTestCase, hopper, unittest

try:
    from PIL import _webp
except ImportError:
    _webp = None


@unittest.skipIf(_webp is None, "WebP support not installed")
class TestFileWebpAlpha(PillowTestCase):
    def setUp(self):
        if _webp.WebPDecoderBuggyAlpha(self):
            self.skipTest(
                "Buggy early version of WebP installed, not testing transparency"
            )

    def test_read_rgba(self):
        """
        Can we read an RGBA mode file without error?
        Does it have the bits we expect?
        """

        # Generated with `cwebp transparent.png -o transparent.webp`
        file_path = "Tests/images/transparent.webp"
        image = Image.open(file_path)

        self.assertEqual(image.mode, "RGBA")
        self.assertEqual(image.size, (200, 150))
        self.assertEqual(image.format, "WEBP")
        image.load()
        image.getdata()

        image.tobytes()

        target = Image.open("Tests/images/transparent.png")
        self.assert_image_similar(image, target, 20.0)

    def test_write_lossless_rgb(self):
        """
        Can we write an RGBA mode file with lossless compression without
        error? Does it have the bits we expect?
        """

        temp_file = self.tempfile("temp.webp")
        # temp_file = "temp.webp"

        pil_image = hopper("RGBA")

        mask = Image.new("RGBA", (64, 64), (128, 128, 128, 128))
        # Add some partially transparent bits:
        pil_image.paste(mask, (0, 0), mask)

        pil_image.save(temp_file, lossless=True)

        image = Image.open(temp_file)
        image.load()

        self.assertEqual(image.mode, "RGBA")
        self.assertEqual(image.size, pil_image.size)
        self.assertEqual(image.format, "WEBP")
        image.load()
        image.getdata()

        self.assert_image_equal(image, pil_image)

    def test_write_rgba(self):
        """
        Can we write a RGBA mode file to webp without error.
        Does it have the bits we expect?
        """

        temp_file = self.tempfile("temp.webp")

        pil_image = Image.new("RGBA", (10, 10), (255, 0, 0, 20))
        pil_image.save(temp_file)

        if _webp.WebPDecoderBuggyAlpha(self):
            return

        image = Image.open(temp_file)
        image.load()

        self.assertEqual(image.mode, "RGBA")
        self.assertEqual(image.size, (10, 10))
        self.assertEqual(image.format, "WEBP")
        image.load()
        image.getdata()

        # early versions of webp are known to produce higher deviations:
        # deal with it
        if _webp.WebPDecoderVersion(self) <= 0x201:
            self.assert_image_similar(image, pil_image, 3.0)
        else:
            self.assert_image_similar(image, pil_image, 1.0)

    def test_write_unsupported_mode_PA(self):
        """
        Saving a palette-based file with transparency to WebP format
        should work, and be similar to the original file.
        """

        temp_file = self.tempfile("temp.webp")
        file_path = "Tests/images/transparent.gif"
        Image.open(file_path).save(temp_file)
        image = Image.open(temp_file)

        self.assertEqual(image.mode, "RGBA")
        self.assertEqual(image.size, (200, 150))
        self.assertEqual(image.format, "WEBP")

        image.load()
        image.getdata()
        target = Image.open(file_path).convert("RGBA")

        self.assert_image_similar(image, target, 25.0)
